from obgraph import Graph
from obgraph.variants import VcfVariants, VcfVariant

def test_find_insertion_nodes():
    g = Graph.from_dicts(
        {
            1: "CTACCA",
            2: "AA",
            3: "TAAATAA",
            4: ""
        },
        {
            1: [2, 4],
            2: [3],
            4: [3]

        },
        [1, 3],
        chromosome_start_nodes={1: 1}
    )
    variant = VcfVariant(1, 6, "A", "AAA", "", "INSERTION")
    ref_node, variant_node = g.get_variant_nodes(variant)
    assert ref_node == 4
    assert variant_node == 2


def test_position_index():
    variants = VcfVariants(
        [
            VcfVariant(1, 4),
            VcfVariant(1, 7),
            VcfVariant(1, 8),
            VcfVariant(1, 8),
            VcfVariant(3, 8),
            VcfVariant(3, 9),
            VcfVariant(3, 11),
        ]
    )

    variants.make_position_index()

    v = variants.get_variants_in_region(1, 4, 8)
    print(v)




from collections import defaultdict
import logging


class MutableGraph:
    def __init__(self, nodes=None, node_sequences=None, edges=None, linear_ref_nodes=None, node_to_ref_offset=None, ref_offset_to_node=None, chromosome_start_nodes=None, allele_frequencies=None):
        if nodes is None:
            nodes = {}
        if node_sequences is None:
            node_sequences = {}
        if edges is None:
            edges = {}
        if linear_ref_nodes is None:
            linear_ref_nodes = []


        self.nodes = nodes
        self.node_sequences = node_sequences
        self.edges = edges
        self.linear_ref_nodes = linear_ref_nodes
        if linear_ref_nodes is not None:
            self.linear_ref_nodes_set = set(linear_ref_nodes)
        else:
            self.linear_ref_nodes_set = set()

        self.node_to_ref_offset = node_to_ref_offset
        self.ref_offset_to_node = ref_offset_to_node
        self.chromosome_start_nodes = chromosome_start_nodes
        self.allele_frequencies = allele_frequencies
        self.reverse_edges = defaultdict(list)
        self.get_reverse_edges()

    def __str__(self):
        description = "Nodes: %s " % (self.node_sequences)
        description += "\nEdges: %s" % (self.edges)
        return description

    def get_reverse_edges(self):
        for node in self.get_all_nodes():
            for edge in self.get_edges(node):
                self.reverse_edges[edge].append(node)

    def set_linear_ref_nodes(self, nodes):
        self.linear_ref_nodes = nodes
        self.linear_ref_nodes_set = set(nodes)

    def get_all_nodes(self):
        return list(self.nodes.keys())

    def get_node_size(self, node):
        return self.nodes[node]

    def get_node_sequence(self, node):
        return self.node_sequences[node]

    def get_edges(self,node):
        if node in self.edges:
            return self.edges[node]
        else:
            return []

    def remove_edge(self, from_node, to_node):
        self.edges[from_node].remove(to_node)
        self.reverse_edges[to_node].remove(from_node)

    def add_edge(self, from_node, to_node):
        if from_node not in self.edges:
            self.edges[from_node] = []

        self.edges[from_node].append(to_node)

        if to_node not in self.reverse_edges:
            self.reverse_edges[to_node] = []

        self.reverse_edges[to_node].append(from_node)

    def add_node(self, id, sequence="", is_ref_node=False):
        self.nodes[id] = len(sequence)
        self.node_sequences[id] = sequence

        if is_ref_node:
            self.linear_ref_nodes.append(id)
            self.linear_ref_nodes_set.add(id)

    def get_nodes_before(self, node):
        return self.reverse_edges[node]

    def find_nodes_from_node_that_matches_sequence(self, from_node, sequence, variant_type, nodes_found, all_paths_found):
        #logging.info("== Searching from node %d with sequence %s. Variant type %s. Nodes found: %s. All paths found: %s" % (from_node, sequence, variant_type, nodes_found, all_paths_found))
        if sequence == "":
            # All of the sequence is used successfully, return
            all_paths_found.append(nodes_found)
            #logging.info("   RETURNING. All paths found: %s" % all_paths_found)
            return nodes_found, all_paths_found

        next_nodes = self.get_edges(from_node)
        result = (nodes_found, all_paths_found)
        for possible_next in next_nodes:

            #print("Checking next node %d" % possible_next)

            node_size = self.get_node_size(possible_next)
            if node_size == 0 or self.get_node_sequence(possible_next).lower() == sequence[0:node_size].lower():
                # This node is a match, we can continue searching
                new_sequence = sequence[node_size:]
                new_nodes_found = nodes_found.copy()
                new_nodes_found.append(possible_next)

                #print("   Matching sequence. New sequence is now %s" % new_sequence)
                result = MutableGraph.find_nodes_from_node_that_matches_sequence(self, possible_next, new_sequence, variant_type, new_nodes_found, all_paths_found)
                if not result:
                    continue
            else:
                #print("   No sequence match")
                continue

        return result




from obgraph import Graph, DummyNodeAdder
from obgraph.graph_construction import GraphConstructor
from obgraph.variants import VcfVariants, VcfVariant

def test_simple_insertion():
    graph = Graph.from_dicts(
        {1: "ACTG", 2: "C", 3: "AAAA"},
        {1: [2, 3], 2: [3]},
        [1, 3]
    )

    variants = VcfVariants([VcfVariant(1, 4, "G", "GC", type="INSERTION")])
    dummy_node_adder = DummyNodeAdder(graph, variants)
    new_graph = dummy_node_adder.create_new_graph_with_dummy_nodes()

    assert new_graph.node_has_edges(5, [3])
    assert new_graph.node_has_edges(1, [2, 5])
    assert new_graph.node_has_edges(2, [3])
    #print(new_graph)


def test_double_deletion_with_snp_inside_first_deletion():

    graph = Graph.from_dicts(
        {1: "ACTG", 2: "A", 3: "C", 4: "T", 5: "AAA", 6: "G"},
        {
            1: [2, 5, 6],
            2: [3, 4],
            3: [5, 6],
            4: [5, 6],
            5: [6]
        },
        [1, 2, 4, 6]
    )

    variants = VcfVariants([VcfVariant(1, 4, "GAT", "G", type="DELETION"), VcfVariant(1, 6, "TAAA", "T", type="DELETION")])
    dummy_node_adder = DummyNodeAdder(graph, variants)
    new_graph = dummy_node_adder.create_new_graph_with_dummy_nodes()
    print(new_graph)

def test_double_deletion_with_snp_inside_first_deletiod_and_false_deletion_path():

    repeated_sequence = "AGGTCCCAGGTCCATCT"
    graph = Graph.from_dicts(
        {1: "TTTT", 2: "AGGTCC", 3: "C", 4: "A", 5: repeated_sequence, 6: repeated_sequence},
        {
            1: [2, 5, 6],
            2: [3, 4],
            3: [5, 6],
            4: [5, 6],
            5: [6]
        },
        [1, 2, 3, 5, 6]
    )

    variants = VcfVariants([VcfVariant(1, 4, "TAGGTCCC", "T", type="DELETION"), VcfVariant(1, 11, "CAGGTCCCAGGTCCATCT", "C", type="DELETION")])
    dummy_node_adder = DummyNodeAdder(graph, variants)
    new_graph = dummy_node_adder.create_new_graph_with_dummy_nodes()

    print(new_graph)
    assert list(new_graph.get_edges(1)) == [2, 8]
    assert list(new_graph.get_edges(2)) == [3, 4]
    assert list(new_graph.get_edges(3)) == [5, 9]
    assert list(new_graph.get_edges(4)) == [5, 9]
    assert list(new_graph.get_edges(9)) == [6]
    assert list(new_graph.get_edges(8)) == [5, 9]

def test_insertion_with_multiple_paths():

    graph = Graph.from_dicts(
        {1: "AAAG", 2: "GAGT", 3: "GA", 4: "C", 5: "G", 6: "T"},
        {
            1: [2, 3],
            2: [3],
            3: [4, 5],
            4: [6],
            5: [6]
        },
        [1, 3, 5, 6]
    )

    variants = VcfVariants([VcfVariant(1, 4, "G", "GGAGT", type="INSERTION")])
    dummy_node_adder = DummyNodeAdder(graph, variants)
    new_graph = dummy_node_adder.create_new_graph_with_dummy_nodes()
    assert list(new_graph.get_edges(1)) == [2, 8]
    assert list(new_graph.get_edges(8)) == [3]
    print(new_graph)


def test_tricky_case_nested_deletions():
    graph = Graph.from_dicts(
        {1: "TATAT", 2: "AT", 3: "A", 4: "T", 5: "A", 6: "A", 7: "T", 8: "A", 9: "GG" },
        {
            1: [2, 6],
            2: [3, 6],
            3: [4, 5],
            4: [6],
            5: [6],
            6: [7, 8],
            7: [9],
            8: [9]
        },
        [1, 2, 3, 5, 6, 8, 9]
    )

    variants = VcfVariants(
        [
            VcfVariant(1, 5, "TATAA", "T", type="DELETION"),
            VcfVariant(1, 7, "TAA", "T", type="DELETION"),
            VcfVariant(1, 5, "A", "T", type="SNP"),
        ]
    )

    dummy_node_adder = DummyNodeAdder(graph, variants)
    new_graph = dummy_node_adder.create_new_graph_with_dummy_nodes()
    print(new_graph)

    assert list(new_graph.get_edges(1)) == [2, 11]
    assert list(new_graph.get_edges(2)) == [3, 12]
    assert list(new_graph.get_edges(11)) == [6]
    assert list(new_graph.get_edges(12)) == [6]



def test_overlapping_deletions():
    graph = Graph.from_dicts(
        {1: "AA", 2: "TCTG", 3: "TCT", 4: "G", 5: "A", 6: "GG"},
        {
            1: [2, 3],
            2: [3, 6],
            3: [4, 5],
            4: [6],
            5: [6]
        },
        [1, 2, 3, 5, 6]
    )

    variants = VcfVariants(
        [
            VcfVariant(1, 2, "ATCTG", "A", type="DELETION"),
            VcfVariant(1, 6, "GTCTA", "T", type="DELETION"),
            VcfVariant(1, 10, "A", "G", type="SNP")
        ]
    )
    dummy_node_adder = DummyNodeAdder(graph, variants)
    new_graph = dummy_node_adder.create_new_graph_with_dummy_nodes()

    assert list(new_graph.get_edges(1)) == [2, 8]
    assert list(new_graph.get_edges(8)) == [3, 9]
    assert list(new_graph.get_edges(2)) == [3, 9]
    assert list(new_graph.get_edges(9)) == [6]

    ref_node, var_node = new_graph.get_variant_nodes(variants[1])
    assert ref_node == 3
    assert var_node == 9
    print(new_graph)



def test_insertion_with_identical_false_path():
    graph = Graph.from_dicts(
        {1: "AA", 2: "TCTG", 3: "TCTG", 4: "GG"},
        {
            1: [2, 3],
            2: [3],
            3: [4],
        },
        [1, 3, 4]
    )

    variants = VcfVariants(
        [
            VcfVariant(1, 2, "A", "ATCTG", type="INSERTION"),
        ]
    )
    dummy_node_adder = DummyNodeAdder(graph, variants)
    new_graph = dummy_node_adder.create_new_graph_with_dummy_nodes()

    print(new_graph)

    assert list(new_graph.get_edges(1)) == [2, 6]
    assert list(new_graph.get_edges(6)) == [3]
    assert list(new_graph.get_edges(2)) == [3]

    ref_node, var_node = new_graph.get_variant_nodes(variants[0])
    assert ref_node == 6
    assert var_node == 2


def test_full_create_with_two_insertions_at_same_position():
    variants = VcfVariants(
        [
            VcfVariant(1, 3, "A", "ATATT", type="INSERTION"),
            VcfVariant(1, 3, "A", "ATATTTATT", type="INSERTION")
        ]
    )

    ref_sequence = "ATAGGGGGG"
    constructor = GraphConstructor(ref_sequence, variants)
    without_dummy = constructor._graph.to_mutable_graph()
    print(without_dummy)
    print(without_dummy.nodes)

    assert set(without_dummy.edges[1]) == set([2, 3, 4])
    assert set(without_dummy.edges[2]) == set([4])
    assert set(without_dummy.edges[3]) == set([4])

    graph = constructor.get_graph_with_dummy_nodes()
    assert set(graph.edges[1]) == set([2, 3, 6, 7])
    assert set(graph.edges[2]) == set([4])
    assert set(graph.edges[6]) == set([4])


def test_two_overlapping_snps():
    variants = VcfVariants(
        [
            VcfVariant(1, 3, "A", "G", type="SNP"),
            VcfVariant(1, 3, "A", "T", type="SNP"),
        ]
    )

    ref_sequence = "CCACCC"
    constructor = GraphConstructor(ref_sequence, variants)
    without_dummy = constructor._graph.to_mutable_graph()
    print(without_dummy)
    print(without_dummy.nodes)

    graph = constructor.get_graph_with_dummy_nodes()

    correct = Graph.from_dicts({
            1: "CC",
            2: "G",
            3: "T",
            4: "A",
            5: "CCC"
        },
        {
            1: [2, 3, 4],
            2: [5],
            3: [5],
            4: [5]
        },
        [1, 4, 5])

    assert correct == graph


def test_two_overlapping_snps_and_indel():
    variants = VcfVariants(
        [
            VcfVariant(1, 3, "A", "G", type="SNP"),
            VcfVariant(1, 3, "A", "T", type="SNP"),
            VcfVariant(1, 3, "A", "GCT", type="INSERTION")
        ]
    )

    ref_sequence = "CCACCC"
    constructor = GraphConstructor(ref_sequence, variants)
    without_dummy = constructor._graph.to_mutable_graph()
    print(without_dummy)
    print(without_dummy.nodes)

    graph = constructor.get_graph_with_dummy_nodes()

    print(graph.to_mutable_graph())
    assert graph.get_node_sequence(8) == ""
    print("Sequence: ", graph.get_node_sequence(8))
    assert graph.get_edges(8) == [6]
    assert all(graph.get_edges(4) == [5, 8])
    assert graph.get_node_sequence(5) == "CT"
    print(graph.get_edges(8))


import logging
from .mutable_graph import MutableGraph
from .graph import Graph
import numpy as np
import pickle


def convert_gfa_ids_to_numeric(gfa_file_name, out_base_name):
    gfa = open(gfa_file_name)
    out_gfa = open(out_base_name + ".gfa", "w")
    id_mapping = {}

    current_id = 0
    for line in gfa:
        l = line.strip().split()
        if l[0] == "S" or l[0] == "A":
            id = l[1]
            if id in id_mapping:
                new_id = str(id_mapping[id])
            else:
                new_id = str(current_id)
                id_mapping[id] = current_id
                current_id += 1

            l[1] = new_id
        elif l[0] == "L":
            assert l[1] in id_mapping and l[3] in id_mapping, "Gfa may have L lines before S lines"
            l[1] = str(id_mapping[l[1]])
            l[3] = str(id_mapping[l[3]])

        new_line = "\t".join(l)
        out_gfa.write(new_line + "\n")

    with open(out_base_name + ".id_mapping", "wb") as f:
        pickle.dump(id_mapping, f)
        logging.info("Wrote id mapping to %s" % out_base_name + ".id_mapping")


def create_graph_from_gfa_file(file_name):

    graph = MutableGraph()

    path_lines = []

    with open(file_name) as f:
        for line in f:
            l = line.split()
            if line.startswith("S"):
                id = int(l[1])
                sequence = l[2]
                graph.add_node(id, sequence)
            elif line.startswith("L"):
                from_node = int(l[1])
                to_node = int(l[3])

                if l[2] != "+" or l[4] != "+":
                    logging.warning("Only links from positive side to positive side are supported. Found link with negative side")

                if l[5] != "*":
                    logging.warning("Overlaps between segments not supported. Segments %d-%d will be linked by edge"
                                    % (from_node, to_node))

                graph.add_edge(from_node, to_node)
            elif line.startswith("P"):
                if not l[1].startswith("_alt"):
                    path_lines.append(l)

    #assert len(path_lines) == 1, "Supports only one path, which should be reference path. Found %d" % len(path_lines)

    # find reference node from path
    linear_ref_nodes = []
    chromosome_start_nodes = {}
    for path_line in path_lines:
        path_name = path_line[1]
        logging.info("Processing path %s" % path_name)
        nodes = path_line[2].split(",")
        nodes = [int(node.replace("+", "")) for node in nodes]
        chromosome_start_nodes[path_name] = nodes[0]
        linear_ref_nodes.extend(nodes)
        logging.info("There are %d linear ref nodes in path %s"  % (len(nodes), path_name))

    graph.set_linear_ref_nodes(linear_ref_nodes)
    graph.chromosome_start_nodes = chromosome_start_nodes

    logging.info("Chromosome start nodes: %s" % graph.chromosome_start_nodes)
    return Graph.from_mutable_graph(graph)





from obgraph.haplotype_nodes import HaplotypeNodes, NodeToHaplotypes, HaplotypeToNodes
import numpy as np
from obgraph.haplotype_nodes import DisckBackedRaggedArray


def test_node_to_haplotypes():
    nodes = np.zeros((2, 10))
    nodes[0,0:3] = np.array([1, 2, 3])
    nodes[1,0:3] = np.array([1, 2, 4])
    haplotype_nodes = HaplotypeNodes(nodes, None)

    lookup = NodeToHaplotypes.from_haplotype_nodes(haplotype_nodes)
    lookup.to_file("testfile")
    lookup2 = NodeToHaplotypes.from_file("testfile")

    assert list(lookup2.get_haplotypes_on_node(1)) == [0, 1]
    assert list(lookup2.get_haplotypes_on_node(2)) == [0, 1]
    assert list(lookup2.get_haplotypes_on_node(3)) == [0]
    assert list(lookup2.get_haplotypes_on_node(4)) == [1]


def test_haplotype_to_nodes():
    haplotype_to_nodes = HaplotypeToNodes.from_flat_haplotypes_and_nodes(
        [0, 0, 0, 4, 4, 2, 2, 2, 2],
        [1, 2, 3, 1, 3, 1, 2, 3, 4]
    )


    haplotype_to_nodes.to_file("testfile")
    new = HaplotypeToNodes.from_file("testfile")

    assert list(new.get_nodes(0)) == [1, 2, 3]
    assert list(new.get_nodes(4)) == [1, 3]
    assert list(new.get_nodes(2)) == [1, 2, 3, 4]



def test_disc_based_ragged_array():
    data = [
        np.arange(10, dtype=np.int64),
        np.array([4, 3, 1], dtype=np.int64),
        np.array([], dtype=np.int64),
        np.array([10], dtype=np.int64),
    ]

    ra = DisckBackedRaggedArray.from_iter("test_ra", data)
    print(ra._offsets, ra._lengths)

    for row, d in enumerate(data):
        print(row, d, ra[row])
        assert np.all(ra[row] == d)


import numpy as np
from obgraph.position_id import PositionId
from obgraph import Graph


class DummyGraph:
    def __init__(self, nodes):
        self.nodes = nodes

def test():
    graph = DummyGraph(np.array([1, 3, 1, 0, 5, 4]))
    position_id = PositionId.from_graph(graph)

    assert position_id.get(0, 0) == 0
    assert position_id.get(1, 0) == 1
    assert position_id.get(1, 1) == 2
    assert position_id.get(1, 2) == 3
    assert position_id.get(2, 0) == 4


def test2():
    graph = Graph.from_dicts(
        {1: "CCCCC", 2: "G", 3: "", 4: "ACT", 5: "", 6: "GC", 7: "A", 8: "T", 9: "G", 10: "GGG"},
        {1: [2, 3], 2: [4], 3: [4], 4: [5, 6], 5: [7], 6: [7], 7: [8, 9], 8: [10], 9: [10]},
        [1, 2, 4, 7, 8, 10]
    )
    position_id = PositionId.from_graph(graph)
    ids = set()
    for node in graph.get_all_nodes():
        if graph.get_node_size(node):
            for offset in range(0, graph.get_node_size(node)):
                id = position_id.get(node, offset)
                assert id not in ids
                ids.add(id)

test()
test2()




import logging

from obgraph.graph_construction import GraphConstructor as GraphConstructor
from obgraph.variants import VcfVariants, VcfVariant


def test_single_snp():
    reference = "ACTGGG"

    variants = VcfVariants(
        [VcfVariant(1, 2, "C", "T", type="SNP")]
    )

    constructor = GraphConstructor(reference, variants)
    graph = constructor.get_graph()

    assert list(graph.get_edges(1)) == [2, 3], graph.get_edges(1)
    assert list(graph.get_edges(2)) == [4], graph.get_edges(2)
    assert list(graph.get_edges(3)) == [4], graph.get_edges(3)
    assert list(graph.get_edges(4)) == [] , graph.get_edges(4)

    assert graph.get_node_size(1) == 1
    assert graph.get_node_size(2) == 1
    assert graph.get_node_size(3) == 1
    assert graph.get_node_size(4) == 4

    assert graph.get_node_sequence(1) == "A"
    assert graph.get_node_sequence(2) == "T"
    assert graph.get_node_sequence(3) == "C"
    assert graph.get_node_sequence(4) == "TGGG"

    assert graph.linear_ref_nodes() == set([1, 3, 4])


def test_deletion_with_snp_at_end():
    reference = "TCTGTCTAGG"
    variants = VcfVariants(
        [VcfVariant(1, 4, "GTCTA", "G", type="DELETION"),
         VcfVariant(1, 8, "A", "T", type="SNP")]
    )
    constructor = GraphConstructor(reference, variants)
    graph = constructor.get_graph()

    print(graph)

def test_single_deletion():
    logging.info("\n\nTest single deletion")
    reference = "AATTGG"

    variants = VcfVariants(
        [VcfVariant(1, 2, "ATT", "A", type="DELETION")]
    )

    constructor = GraphConstructor(reference, variants)
    graph = constructor.get_graph()
    print(graph)


    assert list(graph.get_edges(1)) == [2, 5]
    assert list(graph.get_edges(2)) == [3]

    assert graph.get_node_sequence(5) == ""
    assert graph.get_node_sequence(2) == "TT"
    assert graph.get_node_sequence(1) == "AA"
    assert graph.get_node_sequence(3) == "GG"

    assert graph.linear_ref_nodes() == set([1, 2, 3])



def test_single_insertion():
    reference = "AATTGG"

    variants = VcfVariants(
        [VcfVariant(1, 2, "A", "AAA", type="INSERTION")]
    )

    constructor = GraphConstructor(reference, variants)
    graph = constructor.get_graph_with_dummy_nodes()
    print(graph)

    assert list(graph.get_edges(1)) == [2, 5]
    assert list(graph.get_edges(5)) == [3]
    assert list(graph.get_edges(3)) == []


def test_single_substitution():
    reference = "AATTGG"

    variants = VcfVariants(
        [VcfVariant(1, 2, "ATT", "ACC", type="SUBSTITUTION")]
    )

    constructor = GraphConstructor(reference, variants)
    graph = constructor.get_graph_with_dummy_nodes()
    assert graph.get_node_sequence(1) == "AA"
    assert graph.get_node_sequence(2) == "CC"
    assert graph.get_node_sequence(3) == "TT"
    assert graph.get_node_sequence(4) == "GG"
    assert list(graph.get_edges(1)) == [2, 3]
    assert list(graph.get_edges(2)) == [4]
    assert list(graph.get_edges(3)) == [4]


def test_double_deletion_with_snp_inside_first_deletion():

    reference = "ACTGATAAAG"
    variants = VcfVariants([VcfVariant(1, 4, "GAT", "G", type="DELETION"), VcfVariant(1, 6, "T", "C", type="SNP"), VcfVariant(1, 6, "TAAA", "T", type="DELETION")])
    constructor = GraphConstructor(reference, variants)
    graph = constructor.get_graph()
    print(graph.get_node_at_ref_offset(0))
    print(graph.get_node_at_ref_offset(1))
    graph_with_dummy_nodes = constructor.get_graph_with_dummy_nodes()
    print(graph_with_dummy_nodes)

def test_insertion_with_snp_right_before():
    reference = "AAAAAA"
    variants = VcfVariants([VcfVariant(1, 2, "A", "T", type="SNP"), VcfVariant(1, 2, "A", "AC", type="INSERTION")])
    constructor = GraphConstructor(reference, variants)
    graph = constructor.get_graph_with_dummy_nodes()

    assert list(graph.get_edges(1)) == [2, 3]
    assert list(graph.get_edges(2)) == [4, 7]
    assert list(graph.get_edges(3)) == [4, 7]
    assert list(graph.get_edges(4)) == [5]
    assert list(graph.get_edges(5)) == []
    assert list(graph.get_edges(7)) == [5]
    assert graph.get_node_sequence(5) == "AAAA"
    assert graph.get_node_sequence(7) == ""
    assert graph.get_node_sequence(4) == "C"
    assert graph.get_node_sequence(2) == "T"
    assert graph.linear_ref_nodes() == set([1, 3, 5])


def test_insertion_with_snp_right_before_and_right_after():
    reference = "AAAAAA"
    variants = VcfVariants([VcfVariant(1, 2, "A", "T", type="SNP"), VcfVariant(1, 2, "A", "AC", type="INSERTION"), VcfVariant(1, 3, "A", "C", type="SNP")])
    constructor = GraphConstructor(reference, variants)
    graph = constructor.get_graph()
    print(graph)

def test_deletion_with_snp_right_before_and_right_after():
    reference = "AAAAAA"
    variants = VcfVariants([VcfVariant(1, 2, "A", "T", type="SNP"), VcfVariant(1, 2, "AA", "A", type="DELETION"), VcfVariant(1, 3, "A", "C", type="SNP")])
    constructor = GraphConstructor(reference, variants)
    graph = constructor.get_graph()
    graph_with_dummy_nodes = constructor.get_graph_with_dummy_nodes()
    print(graph)
    print(graph_with_dummy_nodes)


def test_messy_graph():
    reference = "GCATATTTT"
    variants = VcfVariants(
        [
            VcfVariant(1, 2, "CAT", "C", type="DELETION"),
            VcfVariant(1, 3, "A", "G", type="SNP"),
            VcfVariant(1, 4, "TA", "T", type="DELETION"),
            VcfVariant(1, 5, "A", "AT", type="INSERTION"),
        ]
    )

    constructor = GraphConstructor(reference, variants)
    graph = constructor.get_graph()
    print(graph)
    print(constructor.get_graph_with_dummy_nodes())

    ref, var = graph.get_variant_nodes(variants[2])
    assert ref == 5
    assert var == 10

    ref, var = graph.get_variant_nodes(variants[0])
    assert ref == 3
    assert var == 9

    ref, var = graph.get_variant_nodes(variants[1])
    assert ref == 3
    assert var == 2

    ref, var = graph.get_variant_nodes(variants[3])
    assert ref == 11
    assert var == 6

    assert list(graph.get_edges(10)) == [6, 11]
    assert list(graph.get_edges(11)) == [7]
    assert list(graph.get_edges(9)) == [5, 10]


test_single_substitution()
test_insertion_with_snp_right_before()
test_single_snp()
test_single_deletion()
test_single_insertion()
test_double_deletion_with_snp_inside_first_deletion()
test_insertion_with_snp_right_before_and_right_after()
test_deletion_with_snp_right_before_and_right_after()
test_messy_graph()
test_deletion_with_snp_at_end()



import logging
import numpy as np
cimport numpy as np
cimport cython
import time

def fill_zeros_increasingly(np.int64_t[:] array):
    cdef int i = 0
    for i in range(1, len(array)):
        if array[i] == 0:
            array[i] = array[i-1] + 1


@cython.wraparound(False)
def traverse_graph_by_following_nodes(graph, np.uint8_t[:] follow_nodes, split_into_chromosomes=False):
    logging.info("Traversing with cython")

    #assert type(follow_nodes) == np.ndarray

    logging.info("N nodes in follow set: %d" % len(follow_nodes))

    cdef np.ndarray[np.uint32_t] nodes_found = np.zeros(len(graph.nodes), dtype=np.uint32)
    cdef int node_index = 0

    # accessing RaggedArray internal stuff to speed things up later
    cdef np.uint32_t[:] edges = graph.edges.ravel()
    cdef np.int64_t[:] node_to_n_edges = graph.edges._shape.lengths
    cdef np.int64_t[:] node_to_edge_index = graph.edges._shape.starts

    cdef np.uint8_t[:] linear_nodes_index = graph.linear_ref_nodes_and_dummy_nodes_index

    cdef np.uint32_t[:] next_nodes
    cdef int edge_i = 0
    cdef int next_node = -1
    cdef int j


    #linear_ref_nodes = graph.linear_ref_nodes()
    cdef int current_node

    #print(type(current_node))
    chromosome_index_positions = []

    for current_node in graph.chromosome_start_nodes.values():
        if split_into_chromosomes:
            chromosome_index_positions.append(node_index)

        while True:
            #if current_node in follow_nodes:
                #nodes_found.append(current_node)
            nodes_found[node_index] = current_node
            node_index += 1

            #next_nodes = graph.get_edges(current_node)

            edge_i = node_to_edge_index[current_node]
            next_nodes = edges[edge_i:edge_i+node_to_n_edges[current_node]]

            if len(next_nodes) == 0:
                break

            next_node = -1
            if len(next_nodes) == 1:
                next_node = next_nodes[0]
            else:
                for j in range(next_nodes.shape[0]):
                #for potential_next in next_nodes:
                    if follow_nodes[next_nodes[j]] == 1 or (next_node == -1 and linear_nodes_index[next_nodes[j]] == 1):
                        next_node = next_nodes[j]

                assert next_node != -1, "Could not find next node from node %d" % current_node

            current_node = next_node

    #if current_node in follow_nodes:
    #nodes_found.append(current_node)
    #nodes_found[node_index] = current_node
    #node_index += 1

    if split_into_chromosomes:
        return nodes_found[0:node_index], chromosome_index_positions

    return nodes_found[0:node_index]



import numpy as np
from collections import defaultdict
import logging
from .graph import VariantNotFoundException, Graph
import pickle
from multiprocessing import Pool, Process
from shared_memory_wrapper import object_to_shared_memory, object_from_shared_memory, get_shared_pool, close_shared_pool, from_file, to_file
from itertools import repeat
from .traversing import traverse_graph_by_following_nodes
import random
from npstructures import RaggedArray
from .util import phased_genotype_matrix_to_haplotype_matrix
from .nplist import NpList
import bionumpy as bnp
import time
from shared_memory_wrapper.util import parallel_map_reduce, ConcatenateReducer
from .util import log_memory_usage_now
from dataclasses import dataclass



class DisckBackedRaggedArray:
    def __init__(self, file_name, offsets, lengths):
        self._file_name = file_name
        self._offsets = offsets
        self._lengths = lengths

    def __getitem__(self, row: int):
        assert isinstance(row, int)
        out = np.fromfile(self._file_name,
                          offset=self._offsets[row] * 8,  # *8 because bytes
                          count=self._lengths[row],
                          dtype=np.int64)
        return out

    @classmethod
    def from_iter(cls, file_name, data):
        with open(file_name, "wb") as out_file:

            offsets = []
            lengths = []
            offset = 0

            for row_data in data:
                assert row_data.dtype == np.int64
                offsets.append(offset)
                length = len(row_data)
                lengths.append(length)
                offset += length
                #out_file.write(row_data)
                row_data.tofile(out_file)

            return cls(file_name, np.array(offsets, dtype=np.uint32), np.array(lengths, dtype=np.uint32))


class GenotypeToNodes:
    def __init__(self):
        # Could this just be a map from genotype to two haplotypes?
        pass

    @classmethod
    def make_from_n_random_haplotypes(cls, graph, variants, n_haplotypes=10):
        pass


@dataclass
class DiscBackedGenotypeMatrix:
    file_name: str
    n_variants: int
    n_individuals: int

    def get_individual_genotypes(self, individual):
        start = individual * n_variants
        end = start + n_variants
        return np.fromfile(self.file_name, offset=individual*self.n_variants, count=n_variants)


class DiscBackedHaplotypeToNodes:
    def __init__(self, data: DisckBackedRaggedArray):
        self.data = data

    def n_haplotypes(self):
        return len(self.data._offsets)

    def get_nodes(self, haplotype):
        return self.data[haplotype]

    def __getitem__(self, item):
        return self.data[item]

    @classmethod
    def from_file(cls, file_name):
        #return from_file(file_name)
        o = from_file(file_name)
        # hack with file name to get relative path correct
        o.data._file_name = file_name + ".haplotype_nodes"
        return o

    def to_file(self, file_name):
        to_file(self, file_name)

    @classmethod
    def from_phased_genotype_matrix(cls, genotype_matrix, variant_to_nodes, out_file_name):

        n_variants = genotype_matrix.shape[0]
        logging.info("%d variants" % n_variants)


        def get_nodes():
            haplotype_id = 0
            for individual in range(genotype_matrix.shape[1]):
                has_variant = phased_genotype_matrix_to_haplotype_matrix(genotype_matrix[:,individual].reshape(n_variants, 1))
                assert has_variant.shape[1] == 2
                for haplotype in [0, 1]:
                    has_node = np.nonzero(has_variant[:, haplotype])[0]
                    nodes = variant_to_nodes.var_nodes[has_node].astype(np.int64)
                    assert np.max(nodes) <= np.max(variant_to_nodes.var_nodes)
                    assert nodes.dtype == np.int64
                    yield nodes

        ra = DisckBackedRaggedArray.from_iter(out_file_name + ".haplotype_nodes", (nodes for nodes in get_nodes()))
        return cls(ra)
        #return cls(out_file_name, np.array(offsets, dtype=np.uint32), np.array(lengths, dtype=np.uint32))


class HaplotypeToNodesRagged:
    def __init__(self, haplotype_to_nodes: RaggedArray):
        self.haplotype_to_nodes = haplotype_to_nodes

    def to_file(self, file_name):
        to_file(self, file_name)

    def n_haplotypes(self):
        return len(self.haplotype_to_nodes)

    def get_nodes(self, haplotype):
        assert type(haplotype) == int
        return self.haplotype_to_nodes[haplotype]

    def __getitem__(self, item):
        return self.get_nodes(item)

    def get_n_haplotypes_on_nodes_array(self, n_nodes):
        return NotImplemented

    @classmethod
    def from_file(cls, file_name):
        return from_file(file_name)


def get_variant_matrix_as_chunks_with_variant_ids(vcf_file_name, write_to_shared_memory=False):
    file = bnp.open(vcf_file_name, buffer_type=bnp.PhasedVCFMatrixBuffer)
    variant_start_id = 0
    t = time.perf_counter()
    for chunk in file.read_chunks(min_chunk_size=5*10000000):
        #genotypes = object_to_shared_memory(chunk.genotypes)
        #logging.info(chunk.genotypes.dtype)
        #"logging.info("Done writing genotypes to shared memory")
        #logging.info("Done writing genotypes to shared memory. Took %.4 sec" % (time.perf_counter()-t))
        logging.info("Took %3.f sec to read %d reads" % (time.perf_counter()-t, chunk.genotypes.shape[0]))
        if write_to_shared_memory:
            logging.info("Writing genotypes to shared memory")
            genotypes = object_to_shared_memory(chunk.genotypes.raw())
        else:
            genotypes = chunk.genotypes.raw()

        yield variant_start_id, variant_start_id+chunk.genotypes.shape[0], genotypes
        logging.info("Took %3.f sec to yield %d reads" % (time.perf_counter()-t, chunk.genotypes.shape[0]))
        variant_start_id += chunk.genotypes.shape[0]
        t = time.perf_counter()


def make_ragged_haplotype_to_nodes(variant_to_nodes, phased_genotype_matrix, n_threads=4):
    output = []
    t = time.perf_counter()
    logging.info("converting matrix")
    n_haplotypes = phased_genotype_matrix.shape[1] * 2
    logging.info("N haplotypes: %d" % n_haplotypes)
    has_variant = phased_genotype_matrix_to_haplotype_matrix(phased_genotype_matrix)
    phased_genotype_matrix = None  # free memory
    logging.info("Matrix converted, time spent: %.3f" % (time.perf_counter()-t))
    log_memory_usage_now("")
    has_variant = has_variant.T  # transpose to use hack with nonzero
    nonzero = np.nonzero(has_variant)  # last element is variants that individuals have, first element are the individuals
    has_variant = None  # free memory

    logging.info("Size of nonzero: %.2f / %.2f GB" % (nonzero[0].nbytes / 1000000000, nonzero[1].nbytes / 1000000000))
    log_memory_usage_now("")
    logging.info("N nonzero elements: %d" % len(nonzero[0]))
    logging.info("Nonzero done, time spent: %.3f" % (time.perf_counter()-t))
    ragged_data = variant_to_nodes.var_nodes[nonzero[1]]
    logging.info("Size of ragged data: %.3f GB" % (ragged_data.nbytes / 1000000000))
    logging.info("Ragged data made, time spent: %.3f" % (time.perf_counter()-t))
    log_memory_usage_now("")
    ragged_lengths = np.bincount(nonzero[0], minlength=n_haplotypes)
    nonzero = None
    logging.info("Size of ragged lengths: %.3f GB" % (ragged_lengths.nbytes / 1000000000))
    log_memory_usage_now("")
    logging.info("Ragged lengths made, time spent: %.3f" % (time.perf_counter()-t))
    out = HaplotypeToNodesRagged(RaggedArray(ragged_data, ragged_lengths))
    log_memory_usage_now("")
    logging.info("Ragged array made, time spent: %.3f" % (time.perf_counter()-t))
    return out


    for haplotype in range(n_haplotypes):
        if haplotype % 50 == 0:
            logging.info("Haplotype %d" % (haplotype))
        nodes = variant_to_nodes.var_nodes[np.nonzero(has_variant[:, haplotype])[0]]
        output[haplotype] = nodes
        ragged_lengths[len(nodes)]

    logging.info("Spent %.4f sec" % (time.perf_counter() - t))
    return HaplotypeToNodesRagged(RaggedArray(ragged_data, ragged_lengths))

    """
    pool = get_shared_pool(n_threads)

    variant_start_id = 0
    chunks = get_variant_matrix_as_chunks_with_variant_ids(vcf_file_name, True)
    results = parallel_map_reduce(_make_ragged_haplotype_to_nodes_for_variant_chunk,
                                  (n_haplotypes, variant_to_nodes), chunks, ConcatenateReducer(axis=-1), n_threads)
    """

    #for variant_start_id, variant_end_id, chunk in chunks:
    #    output = _make_ragged_haplotype_to_nodes_for_variant_chunk(chunk, n_haplotypes, variant_to_nodes)
    #    results.append(RaggedArray(output))

    return HaplotypeToNodesRagged(results)


def _make_ragged_haplotype_to_nodes_for_variant_chunk(n_haplotypes,
                                                      variant_to_nodes, chunk):
    variant_start_id, variant_end_id, chunk = chunk
    chunk = object_from_shared_memory(chunk)
    t = time.perf_counter()
    output = [NpList(dtype=np.uint32) for _ in range(n_haplotypes)]
    n_variants_in_chunk = chunk.shape[0]
    logging.info("Converting to haplotype matrix")
    has_variant = phased_genotype_matrix_to_haplotype_matrix(chunk)
    variant_nodes_for_chunk = variant_to_nodes.var_nodes[variant_start_id:variant_end_id]
    for haplotype in range(n_haplotypes):
        if haplotype % 500 == 0:
            logging.info("Haplotype %d, variant id %d" % (haplotype, variant_start_id))
        nodes = variant_nodes_for_chunk[np.nonzero(has_variant[:, haplotype])[0]]
        output[haplotype].extend(nodes)
    output = [n.get_nparray() for n in output]
    logging.info("Spent %.4f sec on %d variants" % (time.perf_counter() - t, n_variants_in_chunk))
    return RaggedArray(output)


class HaplotypeToNodes:
    properties = {"_haplotype_to_index", "_haplotype_to_n_nodes", "_nodes"}
    def __init__(self, haplotype_to_index=None, haplotype_to_n_nodes=None, nodes=None):
        self._haplotype_to_index = haplotype_to_index
        self._haplotype_to_n_nodes = haplotype_to_n_nodes
        self._nodes = nodes

    def to_file(self, file_name):
        np.savez(file_name, index=self._haplotype_to_index, n=self._haplotype_to_n_nodes, haplotypes=self._nodes)

    def n_haplotypes(self):
        return len(self._haplotype_to_index)

    def describe(self):
        for i in range(len(self._haplotype_to_index)):
            print("Haplotype %d: %s" % (i, self.get_nodes(i)))

    def get_nodes(self, haplotype):
        assert type(haplotype) == int
        index = self._haplotype_to_index[haplotype]
        n = self._haplotype_to_n_nodes[haplotype]

        if n == 0:
            return np.array([])

        return self._nodes[index:index+n]

    def __getitem__(self, item):
        return self.get_nodes(item)

    def get_n_haplotypes_on_nodes_array(self, n_nodes):
        counts = np.zeros(n_nodes)
        for haplotype in range(len(self._haplotype_to_index)):
            nodes = self.get_nodes(haplotype)
            counts[nodes] += 1

        return counts

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name)
        except FileNotFoundError:
            data = np.load(file_name + ".npz")

        return cls(data["index"], data["n"], data["haplotypes"])

    def get_new_by_traversing_graph(self, graph, n_haplotypes, store_only_variant_nodes=False):
        haplotype_to_index = []
        haplotype_to_n_nodes = []
        nodes = []

        index = 0
        for haplotype in range(n_haplotypes):
            haplotype_to_index.append(index)
            nodes_in_haplotype = self.get_nodes(haplotype)

            # Traverse graph by following these nodes,
            new_nodes = traverse_graph_by_following_nodes(graph, set(nodes_in_haplotype))
            logging.info("Got %d new nodes by traversing graph for haplotype %d" % (len(new_nodes), haplotype))

            nodes.extend(new_nodes)
            index += len(new_nodes)
            haplotype_to_n_nodes.append(len(new_nodes))

        new = HaplotypeToNodes(np.array(haplotype_to_index, dtype=np.uint32), np.array(haplotype_to_n_nodes, dtype=np.uint32), np.array(nodes, dtype=np.uint32))
        print("N nodes: %s" % new._haplotype_to_n_nodes)

        return new

    @classmethod
    def from_flat_haplotypes_and_nodes(cls, haplotypes, nodes):
        assert len(haplotypes) == len(nodes)

        logging.info("Creating numpy arrays from %d nodes" % len(nodes))
        nodes = np.array(nodes, dtype=np.uint32)
        haplotypes = np.array(haplotypes, np.uint16)

        logging.info("Sorting haplotypes and nodes")
        sorting = np.argsort(haplotypes)
        nodes = nodes[sorting]
        haplotypes = haplotypes[sorting]

        # Find positions where nodes change (these are our index entries)
        logging.info("Making index")
        diffs = np.ediff1d(haplotypes, to_begin=1)
        unique_entry_positions = np.nonzero(diffs)[0]
        unique_haplotypes = haplotypes[unique_entry_positions]

        lookup_size = int(np.max(haplotypes)) + 1
        lookup = np.zeros(lookup_size, dtype=np.uint32)
        lookup[unique_haplotypes] = unique_entry_positions
        n_entries = np.ediff1d(unique_entry_positions, to_end=len(haplotypes) - unique_entry_positions[-1])
        print("N entries: %s" % n_entries)
        n_nodes = np.zeros(lookup_size, dtype=np.uint32)
        n_nodes[unique_haplotypes] = n_entries

        return cls(lookup, n_nodes, nodes)

    @staticmethod
    def _multiprocess_wrapper(shared_memory_graph_name, variants, limit_to_n_haplotypes=10):
        graph = object_from_shared_memory(shared_memory_graph_name)
        return HaplotypeToNodes.get_flat_haplotypes_and_nodes_from_graph_and_variants(graph, variants, limit_to_n_haplotypes)

    @classmethod
    def make_from_n_random_haplotypes(cls, graph, variants, n_haplotypes=10, weight_by_allele_frequency=True):
        # Simple way of making "arbitrary" haplotypes, just give every nth variant to every haplotype
        if not weight_by_allele_frequency:
            logging.info("Will not weight by allele frequency, will divide haplotypes equally between ref and var nodes")
        current_haplotype = 0

        haplotype_ids = list(range(n_haplotypes))

        flat_haplotypes = []
        flat_nodes = []
        for i, variant in enumerate(variants):
            if i % 100000 == 0:
                logging.info("%d variants processed" % i)

            try:
                reference_node, variant_node = graph.get_variant_nodes(variant)
            except VariantNotFoundException:
                continue

            # Select number of haplotypes on variant node by allele frequency, always minimum 1
            # never more than n_haplotypes-1 (guaranteeing min 1 on ref and min 1 on alt)
            if weight_by_allele_frequency:
                n_haplotypes_on_variant_node = int(round(min(max(1, graph.get_node_allele_frequency(variant_node) * n_haplotypes), n_haplotypes-1)))
            else:
                assert n_haplotypes % 2 == 0, "Number of haplotypes most be divisible by 2"
                n_haplotypes_on_variant_node = n_haplotypes // 2

            haplotypes_on_variant_node = set(random.sample(haplotype_ids, n_haplotypes_on_variant_node))

            for haplotype in haplotype_ids:
                flat_haplotypes.append(haplotype)
                if haplotype in haplotypes_on_variant_node:
                    flat_nodes.append(variant_node)
                else:
                    flat_nodes.append(reference_node)

            """
            # Give variant node to current haplotype
            flat_haplotypes.append(current_haplotype % n_haplotypes)
            flat_nodes.append(variant_node)

            # Give ref node to all others
            for haplotype in range(n_haplotypes):
                if haplotype != current_haplotype  % n_haplotypes:
                    flat_haplotypes.append(haplotype)
                    flat_nodes.append(reference_node)

            current_haplotype += 1
            """

        return cls.from_flat_haplotypes_and_nodes(flat_haplotypes, flat_nodes)

    @staticmethod
    def get_flat_haplotypes_and_nodes_from_graph_and_variants(graph, variants, limit_to_n_haplotypes=10):
        logging.info("Processing %d variants" % len(variants))
        flat_haplotypes = []
        flat_nodes = []
        haplotypes = list(range(0, limit_to_n_haplotypes))
        for i, variant in enumerate(variants):
            if i % 1000000 == 0:
                logging.info("%d variants processed" % i)

            try:
                reference_node, variant_node = graph.get_variant_nodes(variant)
            except VariantNotFoundException:
                continue

            genotypes = variant.vcf_line.split()[9:]
            for haplotype in haplotypes:
                individual_number = haplotype // 2
                assert individual_number < len(genotypes)
                haplotype_number = haplotype - individual_number * 2
                assert haplotype_number == 0 or haplotype_number == 1
                genotype_string = genotypes[individual_number].replace("/", "|")
                if genotype_string == ".":
                    continue

                try:
                    haplotype_string = genotype_string.split("|")[haplotype_number]
                except IndexError:
                    logging.info("genotype string: %s" % genotype_string)
                    logging.error("Could not find haplotype %d for individual %d" % (haplotype_number, individual_number))
                    logging.error("Genotypes: %s" % genotypes)
                    logging.info("N genotypes: %d " % len(genotypes))
                    raise

                if haplotype_string == "1":
                    # Follows the variant, add variant node here. Do not store reference node in order to svae space
                    flat_haplotypes.append(haplotype)
                    flat_nodes.append(variant_node)
                #else:
                #flat_nodes.append(reference_node)

        return flat_haplotypes, flat_nodes

    @classmethod
    def from_graph_and_variants(cls, graph, variants, limit_to_n_haplotypes=10, n_threads=10):
        # Flat structures used to make the index later
        flat_nodes = []
        flat_haplotypes = []

        logging.info("Making pool")
        #pool = Pool(n_threads)
        pool = get_shared_pool(n_threads)
        logging.info("Made pool")
        shared_memory_graph_name = object_to_shared_memory(graph)
        logging.info("Put graph in shared memory")

        i = 0
        for haplotypes, nodes in pool.starmap(HaplotypeToNodes._multiprocess_wrapper, zip(repeat(shared_memory_graph_name), variants.get_chunks(chunk_size=1000), repeat(limit_to_n_haplotypes))):
            logging.info("Done with %d iterations" % i)
            i += 1
            flat_haplotypes.extend(haplotypes)
            flat_nodes.extend(nodes)
            logging.info("Added nodes and haplotypes")

        logging.info("Done processing all variants")

        close_shared_pool()
        return cls.from_flat_haplotypes_and_nodes(flat_haplotypes, flat_nodes)


# Simple placeholder class for representing a matrix
# rows are haplotypes
# columns contain all nodes covered by that haplotype
class NodeToHaplotypes:
    def __init__(self, nodes_to_index, nodes_to_n_haplotypes, haplotypes):
        self._nodes_to_index = nodes_to_index
        self._nodes_to_n_haplotypes = nodes_to_n_haplotypes
        self._haplotypes = haplotypes
        logging.info("Finding n haplotypes (could be slow?)")
        self._n_haplotypes = np.max(haplotypes)
        logging.info("Done finding n haplotypes")

    def get_haplotypes_on_node(self, node):
        if node >= len(self._nodes_to_index):
            return np.array([])

        index_pos = self._nodes_to_index[node]
        n = self._nodes_to_n_haplotypes[node]
        if n == 0:
            return np.array([])

        return self._haplotypes[index_pos:index_pos+n]

    def get_individuals_having_node_pair(self, node1, node2):
        node1_haplotypes = set(self.get_haplotypes_on_node(node1))
        node2_haplotypes = set(self.get_haplotypes_on_node(node2))

        individuals = set()
        for individual_id in range(self._n_haplotypes // 2):
            haplotype1 = individual_id * 2
            haplotype2 = haplotype1 + 1

            if (haplotype1 in node1_haplotypes and haplotype2 in node2_haplotypes) or (haplotype2 in node1_haplotypes and haplotype1 in node2_haplotypes):
                individuals.add(individual_id)

        return individuals


    @classmethod
    def from_haplotype_nodes(cls, haplotype_nodes):

        # "flat" lists of nodes and corresponding haplotypes having those nodes
        nodes = []
        haplotypes = []

        n_haplotypes = haplotype_nodes.nodes.shape[0]
        for haplotype in range(n_haplotypes):
            logging.info("Processing haplotype %d" % haplotype)
            for node in haplotype_nodes.nodes[haplotype, :]:
                if node > 0:
                    nodes.append(node)
                    haplotypes.append(haplotype)

        nodes = np.array(nodes, dtype=np.uint32)
        haplotypes = np.array(haplotypes, np.uint16)

        sorting = np.argsort(nodes)
        nodes = nodes[sorting]
        haplotypes = haplotypes[sorting]

        # Find positions where nodes change (these are our index entries)
        diffs = np.ediff1d(nodes, to_begin=1)
        unique_entry_positions = np.nonzero(diffs)[0]
        unique_nodes = nodes[unique_entry_positions]

        lookup_size = int(np.max(nodes)) + 1
        lookup = np.zeros(lookup_size, dtype=int)
        lookup[unique_nodes] = unique_entry_positions
        n_entries = np.ediff1d(unique_entry_positions, to_end=len(nodes) - unique_entry_positions[-1])
        n_haplotypes = np.zeros(lookup_size, dtype=np.uint16)
        n_haplotypes[unique_nodes] = n_entries

        return cls(lookup, n_haplotypes, haplotypes)

    def to_file(self, file_name):
        np.savez(file_name, index=self._nodes_to_index, n=self._nodes_to_n_haplotypes, haplotypes=self._haplotypes)

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name)
        except FileNotFoundError:
            data = np.load(file_name + ".npz")

        return cls(data["index"], data["n"], data["haplotypes"])




class HaplotypeNodes:
    def __init__(self, nodes, n_haplotypes_on_node):
        self.nodes = nodes
        self.n_haplotypes_on_node = n_haplotypes_on_node


    def to_file(self, file_name):
        np.savez(file_name, nodes=self.nodes, n_haplotypes_on_node=self.n_haplotypes_on_node)

    def __get__(self, item):
        return self.nodes[item]

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name)
        except FileNotFoundError:
            data = np.load(file_name + ".npz")

        return cls(data["nodes"], data["n_haplotypes_on_node"])

    @classmethod
    def from_graph_and_variants(cls, graph, variants, limit_to_n_haplotypes=10):

        # First find all variant nodes that the haplotype has
        haplotypes = list(range(0, limit_to_n_haplotypes))
        variant_nodes_in_haplotype = defaultdict(set)
        for i, variant in enumerate(variants):
            if i % 1000 == 0:
                logging.info("%d variants processed" % i)

            try:
                reference_node, variant_node = graph.get_variant_nodes(variant)
            except VariantNotFoundException:
                continue

            if variant.position == 4871514:
                logging.info("Variant 4871514 has nodes %d/%d" % (reference_node, variant_node))

            genotypes = variant.vcf_line.split()[9:]
            for haplotype in haplotypes:
                individual_number = haplotype // 2
                haplotype_number = haplotype - individual_number * 2
                haplotype_string = genotypes[individual_number].replace("/", "|").split("|")[haplotype_number]
                if haplotype_string == "1":
                    # Follows the variant, add variant node here
                    variant_nodes_in_haplotype[haplotype].add(variant_node)
                else:
                    variant_nodes_in_haplotype[haplotype].add(reference_node)

        # Iterate graph
        logging.info("Iterating graph for each haplotype")
        nodes = np.zeros((len(haplotypes), len(graph.nodes)), dtype=np.uint32)
        n_haplotypes_on_node = np.zeros(len(graph.nodes) + 1, dtype=np.uint32)

        for haplotype in haplotypes:
            logging.info("Handling haplotype %d" % haplotype)
            current_node = graph.get_first_node()
            i = 0
            while True:
                nodes[haplotype, i] = current_node
                n_haplotypes_on_node[current_node] += 1

                next_nodes = graph.get_edges(current_node)
                if len(next_nodes) == 0:
                    break

                next_node = None
                if len(next_nodes) == 1:
                    next_node = next_nodes[0]
                else:
                    for potential_next in next_nodes:
                        if potential_next in variant_nodes_in_haplotype[haplotype]:
                            next_node = potential_next

                if next_node is None:
                    logging.error("Could not find next node from node %d" % current_node)
                    logging.error("Possible next nodes are %s" % next_nodes)
                    raise Exception("")

                current_node = next_node
                i += 1

            nodes[haplotype, i] = current_node
            n_haplotypes_on_node[current_node] += 1

        return cls(nodes, n_haplotypes_on_node)



from obgraph import MutableGraph

def test_create():
    graph = MutableGraph({1: 4, 2: 3, 3: 1, 4: 1}, {1: "ACTG", 2: "A", 3: "C", 4: "AAAA"}, {1: [2, 3], 3: [4], 2: [4]}, [1, 2, 4])

    assert graph.get_node_size(1) == 4
    assert graph.get_edges(1) == [2, 3]
    assert graph.get_nodes_before(2) == [1]
    assert 2 in graph.get_nodes_before(4)
    assert 3 in graph.get_nodes_before(4)
    assert graph.get_node_sequence(2) == "A"


def test_get_nodes_matching_sequence_single_node():
    graph = MutableGraph({1: 4, 2: 3, 3: 1, 4: 1}, {1: "ACTG", 2: "A", 3: "C", 4: "AAAA"}, {1: [2, 3], 3: [4], 2: [4]}, )

    nodes, all_paths = graph.find_nodes_from_node_that_matches_sequence(1, "A", None, [], [])
    print(all_paths)
    assert nodes == [2]

def test_get_nodes_matching_sequence_double_deletion_and_snp():
    graph = MutableGraph(
        {1: 4, 2: 1, 3: 1, 4: 1, 5: 3, 6: 1},
        {1: "ACTG", 2: "A", 3: "C", 4: "T", 5: "AAA", 6: "G"},
        {
            1: [2, 5, 6],
            2: [3, 4],
            3: [5, 6],
            4: [5, 6],
            5: [6]
        }
    )

    nodes, all_paths = graph.find_nodes_from_node_that_matches_sequence(1, "AT", None, [], [])
    print(all_paths)
    assert nodes == [2, 4]

    nodes, all_paths = graph.find_nodes_from_node_that_matches_sequence(1, "AAA", None, [], [])
    print(all_paths)
    assert nodes == [5]

def test_get_nodes_matching_multiple_paths():
    graph = MutableGraph(
        {1: 1, 2: 1, 3: 1, 4: 1, 5: 2},
        {1: "A", 2: "G", 3: "G", 4: "A", 5: "GG"},
        {
            1: [2, 4, 5],
            2: [3],
            3: [5],
            4: [5],
        }
    )

    nodes, all_paths = graph.find_nodes_from_node_that_matches_sequence(1, "GG", "", [], [])
    print(all_paths)
    assert len(all_paths) == 2
    assert [2, 3] in all_paths
    assert [5] in all_paths


test_create()
test_get_nodes_matching_sequence_single_node()
test_get_nodes_matching_sequence_double_deletion_and_snp()
test_get_nodes_matching_multiple_paths()



import logging
logging.basicConfig(level=logging.INFO)
from obgraph.gfa import create_graph_from_gfa_file


def test_graph_from_gfa():
    gfa_lines = [
        "S 1 ACT",
        "S 2 G",
        "S 3 GGGG",
        "S 4 AAAA",
        "L 1 + 2 + *",
        "L 1 + 3 + *",
        "L 2 + 4 + *",
        "L 3 + 4 + *",
        "P 1 1+,2+,4+ *"
    ]

    with open("gfa.tmp", "w") as f:
        f.writelines((l + "\n" for l in gfa_lines))

    graph = create_graph_from_gfa_file("gfa.tmp")

    assert list(graph.get_edges(1)) == [2, 3]
    assert list(graph.get_edges(2)) == [4]
    assert list(graph.get_edges(3)) == [4]
    assert list(graph.get_edges(4)) == []
    assert graph.get_node_sequence(3) == "GGGG"
    assert graph.get_node_sequence(4) == "AAAA"

    assert graph.is_linear_ref_node(1)
    assert not graph.is_linear_ref_node(3)
    assert graph.is_linear_ref_node(2)
    assert graph.is_linear_ref_node(4)


test_graph_from_gfa()


import logging
import numpy as np
from .graph import Graph


def merge_graphs(graphs):

    # Makes list of all data, then concatenates those in the end into numpy arrays that can be used to create a new graph
    new_nodes = []
    new_node_sequences = []
    new_edges = []
    new_node_to_ref_offset = []
    new_ref_offset_to_node = []
    new_chromosome_start_nodes = {}
    new_allele_frequencies = []

    node_offset = 0
    node_sequence_offset = 0
    edge_index_offset = 0
    ref_offset = 0

    for i, graph in enumerate(graphs):
        logging.info("Processing graph %i" % i)
        logging.info("Node offset is now %d" % node_offset)

        # Add nodes, these are simply just appended
        new_nodes.append(graph.nodes)

        # All node ids should be increased by node_offset

        # RaggedArrays can be concatenated
        new_node_sequences.append(graph.sequences)
        new_edges.append(graph.edges+node_offset)

        new_node_to_ref_offset.append(graph.node_to_ref_offset+ref_offset)
        new_ref_offset_to_node.append(graph.ref_offset_to_node+node_offset)

        assert len(graph.chromosome_start_nodes) == 1, "Can only merge graphs representing single chromosomes"
        chromosome = list(graph.chromosome_start_nodes.keys())[0]
        new_chromosome_start_nodes[chromosome] = graph.get_first_node() + node_offset
        #new_chromosome_start_nodes.append(graph.get_first_node() + node_offset)

        if graph.allele_frequencies is not None:
            new_allele_frequencies.append(graph.allele_frequencies)

        # Increase offsets
        node_offset += len(graph.nodes)
        node_sequence_offset += len(graph.sequences)
        edge_index_offset += len(graph.edges)

        # increase the length of the linear reference genome
        ref_offset += len(graph.ref_offset_to_node)

        logging.info("Ref offset is now %d" % ref_offset)

    logging.info("Concatenating all data")
    new_nodes = np.concatenate(new_nodes)
    new_node_sequences = np.concatenate(new_node_sequences)
    new_edges = np.concatenate(new_edges)
    new_node_to_ref_offset = np.concatenate(new_node_to_ref_offset)
    new_ref_offset_to_node = np.concatenate(new_ref_offset_to_node)

    if len(new_allele_frequencies) > 0:
        new_allele_frequencies = np.concatenate(new_allele_frequencies)
    else:
        new_allele_frequencies = None

    return Graph(new_nodes, new_node_sequences, new_edges,
                 new_node_to_ref_offset, new_ref_offset_to_node, new_chromosome_start_nodes,
                 new_allele_frequencies)



import numpy as np

class PositionId:
    def __init__(self, index):
        self._index = index

    def get(self, nodes, offsets):
        return self._index[nodes]+offsets

    @classmethod
    def from_graph(cls, graph):
        node_sizes = graph.nodes
        node_ids = np.where(node_sizes!=0)[0]
        index = np.zeros(len(node_sizes)+1, dtype=np.int64)
        index[node_ids[1:]] = np.cumsum(node_sizes)[node_ids[1:]-1]
        index[node_ids[0]] = 0
        return cls(index)



import numpy as np
from .graph import VariantNotFoundException
import logging


class VariantToNodes:
    properties = {"ref_nodes", "var_nodes"}
    def __init__(self, ref_nodes=None, var_nodes=None):
        self.ref_nodes = ref_nodes
        self.var_nodes = var_nodes

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name)
        except FileNotFoundError:
            data = np.load(file_name + ".npz")

        return cls(data["ref_nodes"], data["var_nodes"])

    def to_file(self, file_name):
        np.savez(file_name, ref_nodes=self.ref_nodes, var_nodes=self.var_nodes)

    def slice(self, from_variant, to_variant):
        return VariantToNodes(self.ref_nodes[from_variant:to_variant], self.var_nodes[from_variant:to_variant])

    @classmethod
    def from_graph_and_variants(cls, graph, variants):
        n_variants = len(variants)
        var_nodes = np.zeros(n_variants, dtype=np.uint32)
        ref_nodes = np.zeros(n_variants, dtype=np.uint32)

        max_graph_node = graph.max_node_id()
        for i, variant in enumerate(variants):
            if i % 100000 == 0:
                logging.info("%d variants processed" % i)
            try:
                ref_node, var_node = graph.get_variant_nodes(variant)
            except VariantNotFoundException as e:
                logging.error(str(e))
                logging.error("Could not find variant, aborting")
                raise

            var_nodes[i] = var_node
            ref_nodes[i] = ref_node

            assert var_node <= max_graph_node
            assert ref_node <= max_graph_node, "Found ref node %d which is not <= max graph node %d. Variant %s" % (ref_node, max_graph_node, variant)

        return cls(ref_nodes, var_nodes)


    def __iter__(self):
        return zip(self.ref_nodes, self.var_nodes)

    def len(self):
        return len(self.ref_nodes)


class NodeToVariants:
    properties = {"index"}
    def __init__(self, index):
        self.index = index

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name)
        except FileNotFoundError:
            data = np.load(file_name + ".npz")

        return cls(data["index"])

    def to_file(self, file_name):
        np.savez(file_name, index=self.index)

    def get_variant_at_node(self, node):
        if self.index[node] == -1:
            return None
        return self.index[node]

    @classmethod
    def from_variant_to_nodes(cls, variant_to_nodes):
        n_nodes = max(np.max(variant_to_nodes.ref_nodes), np.max(variant_to_nodes.var_nodes))
        index = np.zeros(n_nodes+1, dtype=np.int32) - 1

        for variant in range(len(variant_to_nodes.ref_nodes+1)):
            if variant % 100000 == 0:
                logging.info("%d variants processed" % variant)

            ref_node = variant_to_nodes.ref_nodes[variant]
            var_node = variant_to_nodes.var_nodes[variant]

            index[ref_node] = variant
            index[var_node] = variant

        return cls(index)



from .graph import Graph, VariantNotFoundException
from .dummy_node_adder import DummyNodeAdder
from .mutable_graph import MutableGraph
from .genotype_matrix import GenotypeFrequencies, MostSimilarVariantLookup


import logging
import pickle

logging.basicConfig(level=logging.INFO, format='%(module)s %(asctime)s %(levelname)s: %(message)s')
from obgraph.cython_traversing import traverse_graph_by_following_nodes
import pyximport; pyximport.install()
import sys
import argparse
from . import Graph
from .util import add_indel_dummy_nodes
from .variants import VcfVariants
from .haplotype_nodes import HaplotypeToNodes, NodeToHaplotypes
from .dummy_node_adder import DummyNodeAdder
from .haplotype_nodes import NodeToHaplotypes
from .genotype_matrix import GenotypeMatrix, GenotypeMatrixAnalyser, GenotypeFrequencies
from pyfaidx import Fasta
from .graph_construction import GraphConstructor
from .graph_merger import merge_graphs
import numpy as np
from shared_memory_wrapper import from_shared_memory, to_shared_memory, SingleSharedArray, remove_shared_memory_in_session, to_file, from_file, get_shared_pool, close_shared_pool
from multiprocessing import Pool
import time
from itertools import repeat
from .util import create_coordinate_map
from .util import fill_zeros_with_last
from .variant_to_nodes import VariantToNodes
from npstructures import RaggedArray
import bionumpy as bnp



def merge_graphs_command(args):
    graphs = [Graph.from_file(graph) for graph in args.graphs]
    logging.info("Done reading graphs")

    merged_graph = merge_graphs(graphs)
    merged_graph.to_file(args.out_file_name)


def _assert_vcf_and_fasta_are_compatible(fasta_file_name, vcf_file_name):
    logging.info(f"Checking that fasta {fasta_file_name} and vcf {vcf_file_name} are compatible")
    genome = bnp.Genome.from_file(fasta_file_name)
    chromosomes = genome.get_genome_context().chrom_sizes.keys()

    for chunk in bnp.open(vcf_file_name).read_chunks():
        unique_chromosomes = set([c.chromosome.to_string() for c in chunk])
        assert all(c in chromosomes for c in unique_chromosomes), \
            "VCF contains chromosomes %s. Some chromosomes are not in fasta which contains chromosomes %s" % (unique_chromosomes, chromosomes)


def _assert_chromosome_is_in_reference(fasta_file_name, chromosome):
    genome = bnp.Genome.from_file(fasta_file_name)
    chromosomes = genome.get_genome_context().chrom_sizes.keys()
    assert chromosome in chromosomes, f"Chromosome {chromosome} is not valid, does not exist in fasta file {fasta_file_name} which contains chromosomes {chromosomes}"


def make(args):
    if args.vcf is not None:
        if args.chromosome is not None:
            _assert_chromosome_is_in_reference(args.reference_fasta_file, args.chromosome)
        else:
            _assert_vcf_and_fasta_are_compatible(args.reference_fasta_file, args.vcf)

        logging.info("Will create from vcf file")
        reference = Fasta(args.reference_fasta_file)

        chromosome = args.chromosome
        """
        numeric_chromosome = chromosome
        
        if chromosome == "X":
            numeric_chromosome = "23"
        elif chromosome == "Y":
            numeric_chromosome = "24"
        """

        ref_sequence = str(reference[args.chromosome])
        logging.info("Extracted sequence for chromosome %s. Length is: %d" % (chromosome, len(ref_sequence)))
        variants = VcfVariants.from_vcf(args.vcf, limit_to_chromosome=chromosome, dont_encode_chromosomes=True)
        logging.info("There are %d variants in chromosome" % len(variants))
        assert len(variants) > 0, "Did not find any variants in VCF when limiting to chromosome %s" % chromosome

        constructor = GraphConstructor(ref_sequence, variants)
        graph = constructor.get_graph_with_dummy_nodes()
        graph.to_file(args.out_file_name)
    else:
        logging.info("Will create from files %s" % args.vg_json_files)
        graph = Graph.from_vg_json_files(args.vg_json_files)
        graph.to_file(args.out_file_name)


def add_indel_nodes(args):
    variants = VcfVariants.from_vcf(args.vcf_file_name)
    graph = Graph.from_file(args.graph_file_name)
    adder = DummyNodeAdder(graph, variants)
    new_graph = adder.create_new_graph_with_dummy_nodes()
    edge_mapping = adder.get_edge_mapping()
    new_graph.to_file(args.out_file_name)
    logging.info("Chromosome start nodes in new graph: %s" % new_graph.chromosome_start_nodes)
    logging.info("Wrote new graph to file %s" % args.out_file_name)
    with open(args.out_file_name + ".edge_mapping", "wb") as f:
        pickle.dump(edge_mapping, f)
        logging.info("Wrote edge mapping from old edges to new dummy nodes to file %s.edge_mapping" % args.out_file_name)

def add_allele_frequencies(args):
    logging.info("Reading graph")
    graph = Graph.from_file(args.graph_file_name)
    variants = VcfVariants.from_vcf(args.vcf_file_name, limit_to_chromosome=args.chromosome, skip_index=True, dont_encode_chromosomes=True)
    graph.set_allele_frequencies_from_variants(variants, use_chromosome=args.chromosome)
    graph.to_file(args.graph_file_name)
    logging.info("Wrote modified graph to the same file %s" % args.graph_file_name)


def make_haplotype_to_nodes(args):
    get_shared_pool(args.n_threads)

    graph = Graph.from_file(args.graph_file_name)
    variants = VcfVariants.from_vcf(args.vcf_file_name, make_generator=True, skip_index=True)
    haplotype_to_nodes = HaplotypeToNodes.from_graph_and_variants(graph, variants, args.n_haplotypes, n_threads=args.n_threads)
    # todo: Option to add dummy haplotypes for "all variant nodes" and "all reference nodes" (two for each, to mimic an individual)

    #haplotype_to_nodes = haplotype_to_nodes.get_new_by_traversing_graph(graph, args.n_haplotypes)
    logging.info("Saving to file")
    haplotype_to_nodes.to_file(args.out_file_name)
    logging.info("Wrote to file %s" % args.out_file_name)


def main():
    run_argument_parser(sys.argv[1:])


def run_argument_parser(args):
    parser = argparse.ArgumentParser(
        description='Obgrapph.',
        prog='obgraph',
        formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=50, width=100))

    subparsers = parser.add_subparsers()
    subparser = subparsers.add_parser("make")
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.add_argument("-j", "--vg-json-files", nargs='+', required=False)
    subparser.add_argument("-v", "--vcf", required=False)
    subparser.add_argument("-r", "--reference_fasta_file", required=False)
    subparser.add_argument("-c", "--chromosome", required=False)
    subparser.set_defaults(func=make)

    subparser = subparsers.add_parser("add_indel_nodes")
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.add_argument("-g", "--graph-file-name", required=True)
    subparser.add_argument("-v", "--vcf-file-name", required=True)
    subparser.set_defaults(func=add_indel_nodes)

    subparser = subparsers.add_parser("add_allele_frequencies")
    subparser.add_argument("-g", "--graph-file-name", required=True)
    subparser.add_argument("-v", "--vcf-file-name", required=True)
    subparser.add_argument("-c", "--chromosome", required=False, help="If vcf contains multiple chromsomes, use this to limit to the chromosome that the graph is made from")
    subparser.set_defaults(func=add_allele_frequencies)

    subparser = subparsers.add_parser("make_haplotype_to_nodes")
    subparser.add_argument("-g", "--graph-file-name", required=True)
    subparser.add_argument("-v", "--vcf-file-name", required=True)
    subparser.add_argument("-n", "--n-haplotypes", type=int, required=True)
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.add_argument("-t", "--n-threads", type=int, default=8, required=False)
    subparser.set_defaults(func=make_haplotype_to_nodes)


    def make_haplotype_to_nodes_bnp(args):
        from .haplotype_nodes import make_ragged_haplotype_to_nodes
        phased_genotype_matrix = from_file(args.phased_genotype_matrix).matrix
        variant_to_nodes = VariantToNodes.from_file(args.variant_to_nodes)

        if args.make_disc_backed:
            from .haplotype_nodes import DiscBackedHaplotypeToNodes
            logging.info("Making disc backed")
            result = DiscBackedHaplotypeToNodes.from_phased_genotype_matrix(phased_genotype_matrix, variant_to_nodes, args.out_file_name)
            result.to_file(args.out_file_name)
        else:
            result = make_ragged_haplotype_to_nodes(variant_to_nodes, phased_genotype_matrix, args.n_threads)
            to_file(result, args.out_file_name)

    subparser = subparsers.add_parser("make_haplotype_to_nodes_bnp")
    subparser.add_argument("-g", "--variant-to-nodes", required=True)
    subparser.add_argument("-v", "--phased-genotype-matrix", required=True)
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.add_argument("-t", "--n-threads", type=int, default=8, required=False)
    subparser.add_argument("-n", "--n-haplotypes", type=int, required=False)
    subparser.add_argument("-d", "--make-disc-backed", type=bool, required=False, default=False, help="Uses less memory")
    subparser.set_defaults(func=make_haplotype_to_nodes_bnp)


    def make_node_to_haplotypes_lookup(args):
        haplotype_nodes = HaplotypeNodes.from_file(args.haplotype_nodes)
        n = NodeToHaplotypes.from_haplotype_nodes(haplotype_nodes)
        n.to_file(args.out_file_name)
        logging.info("Saved to %s" % args.out_file_name)

    subparser = subparsers.add_parser("make_node_to_haplotypes_lookup")
    subparser.add_argument("-H", "--haplotype_nodes", required=True)
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.set_defaults(func=make_node_to_haplotypes_lookup)

    def make_genotype_matrix(args):
        from .genotype_matrix import GenotypeMatrix, PhasedGenotypeMatrix

        if args.make_phased_matrix:
            logging.info("Making phased matrix")
            if args.vcf_file_name.endswith(".txt.gz"):
                logging.info("Input is txt")
                matrix = PhasedGenotypeMatrix.from_txt(args.vcf_file_name, args.n_variants, args.n_individuals)
            else:
                matrix = PhasedGenotypeMatrix.from_vcf(args.vcf_file_name, args.n_variants, args.n_individuals)
            to_file(matrix, args.out_file_name)
            return


        if args.n_variants is not None:
            n_variants = args.n_variants
            assert args.n_individuals is not None, "n_individuals must be specified when n_variants is"
            n_individuals = args.n_individuals
        else:
            logging.info("N variants and individuals will be counted from the vcf file. Can take some time.")
            from .util import get_number_of_variants_and_individuals_from_vcf
            n_variants, n_individuals = get_number_of_variants_and_individuals_from_vcf(args.vcf_file_name)

        variants = VcfVariants.from_vcf(args.vcf_file_name, skip_index=True, limit_to_n_lines=None, make_generator=True, dont_encode_chromosomes=True)

        if args.node_to_haplotypes is not None:
            graph = Graph.from_file(args.graph)
            nodes_to_haplotypes = NodeToHaplotypes.from_file(args.node_to_haplotypes)
            matrix = GenotypeMatrix.from_nodes_to_haplotypes_and_variants(nodes_to_haplotypes, variants, graph, args.n_individuals)
        else:
            logging.info("Making genotype matrix directly from vcf")
            matrix = GenotypeMatrix.from_variants(variants, n_individuals, n_variants, n_threads=args.n_threads, chunk_size=args.chunk_size)

        matrix.to_file(args.out_file_name)


    subparser = subparsers.add_parser("make_genotype_matrix")
    subparser.add_argument("-g", "--graph", required=False)
    subparser.add_argument("-v", "--vcf-file-name", required=True)
    subparser.add_argument("-n", "--n-individuals", type=int, required=False)
    subparser.add_argument("-N", "--node-to-haplotypes", required=False)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.add_argument("-m", "--n-variants", required=False, type=int)
    subparser.add_argument("-t", "--n-threads", required=False, type=int, default=32, help="Number of threads used to fill matrix")
    subparser.add_argument("-c", "--chunk-size", required=False, type=int, default=10000, help="Number of variants to process in each job")
    subparser.add_argument("-p", "--make-phased-matrix", required=False, type=bool, default=False)
    subparser.set_defaults(func=make_genotype_matrix)

    def make_haplotype_matrix(args):
        from .haplotype_matrix import HaplotypeMatrix
        variants = VcfVariants.from_vcf(args.vcf_file_name, skip_index=True, limit_to_n_lines=None,
                                        make_generator=True)
        matrix = HaplotypeMatrix.from_variants(variants, args.n_individuals, args.n_variants,
                                                  n_threads=args.n_threads, chunk_size=args.chunk_size)

        matrix.to_file(args.out_file_name)

    subparser = subparsers.add_parser("make_haplotype_matrix")
    subparser.add_argument("-v", "--vcf-file-name", required=True)
    subparser.add_argument("-n", "--n-individuals", type=int, required=True)
    subparser.add_argument("-N", "--node-to-haplotypes", required=False)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.add_argument("-m", "--n-variants", required=True, type=int)
    subparser.add_argument("-t", "--n-threads", required=False, type=int, default=6,
                           help="Number of threads used to fill matrix")
    subparser.add_argument("-c", "--chunk-size", required=False, type=int, default=10000,
                           help="Number of variants to process in each job")
    subparser.set_defaults(func=make_haplotype_matrix)

    def analyse_genotype_matrix(args):

        whitelist_array = None
        if args.whitelist_array is not None:
            whitelist_array = np.load(args.whitelist_array)

        matrix = GenotypeMatrix.from_file(args.genotype_matrix)
        analyser = GenotypeMatrixAnalyser(matrix, whitelist_array=whitelist_array)
        lookup = analyser.analyse(args.n_threads)
        lookup.to_file(args.out_file_name)
        logging.info("Wrote lookup of most similar genotype to file %s" % args.out_file_name)


    subparser = subparsers.add_parser("analyse_genotype_matrix")
    subparser.add_argument("-G", "--genotype-matrix", required=True)
    subparser.add_argument("-w", "--whitelist-array", required=False, help="Array of whitelist variants")
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.add_argument("-t", "--n-threads", required=False, type=int, help="Number of threads to use", default=8)
    subparser.set_defaults(func=analyse_genotype_matrix)


    def make_transition_probabilities(args):
        from .genotype_matrix import GenotypeTransitionProbabilities, MostSimilarVariantLookup
        probs = GenotypeTransitionProbabilities.from_most_similar_variants_and_matrix(
            MostSimilarVariantLookup.from_file(args.most_similar_variants),
            GenotypeMatrix.from_file(args.genotype_matrix),
            n_threads=args.n_threads
        )
        probs.to_file(args.out_file_name)
        logging.info("Wrote to file %s" % args.out_file_name)

    subparser = subparsers.add_parser("make_transition_probabilities")
    subparser.add_argument("-G", "--genotype-matrix", required=True)
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.add_argument("-m", "--most-similar-variants", required=True)
    subparser.add_argument("-t", "--n-threads", required=False, type=int, help="Number of threads to use", default=8)
    subparser.set_defaults(func=make_transition_probabilities)


    def traverse(args):
        g = Graph.from_file(args.graph)
        haplotype_to_nodes = HaplotypeToNodes.from_file(args.haplotype_nodes)
        #from .traversing import traverse_graph_by_following_nodes

        for haplotype in range(0, args.n_haplotypes):

            nodes_to_follow = np.zeros(len(g.nodes), dtype=np.uint8)
            nodes_to_follow[haplotype_to_nodes.get_nodes(haplotype)] = 1
            start_time = time.time()
            new_nodes = traverse_graph_by_following_nodes(g, nodes_to_follow)
            logging.info("Got %d nodes" % len(new_nodes))
            end_time = time.time()
            logging.info("Time spent on haplotype %d: %.5f" % (haplotype, end_time - start_time))

    subparser = subparsers.add_parser("traverse")
    subparser.add_argument("-g", "--graph", required=True)
    subparser.add_argument("-T", "--type", required=False, default="correct_haplotype_nodes")
    subparser.add_argument("-H", "--haplotype_nodes", required=False)
    subparser.add_argument("-n", "--n_haplotypes", type=int, default=1, required=False)
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.set_defaults(func=traverse)

    def get_genotype_frequencies(args):
        if args.vcf_file is not None:
            logging.warning("Creating naively from af field of vcf file")
            variants = VcfVariants.from_vcf(args.vcf_file)
            frequencies = GenotypeFrequencies.create_naive_from_vcf_af_field(variants)
        else:
            matrix = GenotypeMatrix.from_file(args.genotype_matrix)
            frequencies = GenotypeFrequencies.from_genotype_matrix(matrix, args.n_threads)

        frequencies.to_file(args.out_file_name)
        logging.info("Wrote frequencies to file %s" % args.out_file_name)

    subparser = subparsers.add_parser("get_genotype_frequencies")
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.add_argument("-g", "--genotype-matrix", required=False)
    subparser.add_argument("-v", "--vcf-file", required=False, help="If specified, a naive approach will be used, computing from the AF field")
    subparser.add_argument("-t", "--n-threads", default=10, type=int, required=False)
    subparser.set_defaults(func=get_genotype_frequencies)

    def make_random_haplotypes(args):
        graph = Graph.from_file(args.graph)
        variants = VcfVariants.from_vcf(args.vcf_file_name, skip_index=True)
        haplotype_nodes = HaplotypeToNodes.make_from_n_random_haplotypes(graph, variants, n_haplotypes=args.n_haplotypes, weight_by_allele_frequency=not args.no_frequency_weighting)
        logging.info("Making new haplotypenodes by traversing full graph for each haplotype")
        new = haplotype_nodes.get_new_by_traversing_graph(graph, args.n_haplotypes)
        new.to_file(args.out_file_name)
        logging.info("Wrote haplotypenodes to %s" % args.out_file_name)

    subparser = subparsers.add_parser("make_random_haplotypes")
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.add_argument("-g", "--graph", required=True)
    subparser.add_argument("-v", "--vcf-file-name", required=True)
    subparser.add_argument("-n", "--n-haplotypes", type=int, required=False, default=10)
    subparser.add_argument("-e", "--no-frequency-weighting", type=bool, required=False, default=False, help="Set to True to not weight haplotypes by allele frequency")
    subparser.set_defaults(func=make_random_haplotypes)


    def validate_graph(args):
        variants = VcfVariants.from_vcf(args.vcf)
        graph = Graph.from_file(args.graph)

        for i, variant in enumerate(variants):
            if i % 10000 == 0:
                logging.info("%d variants processed" % i)

            ref_node, var_node = graph.get_variant_nodes(variant)

    subparser = subparsers.add_parser("validate_graph")
    subparser.add_argument("-g", "--graph", required=True)
    subparser.add_argument("-v", "--vcf", required=True)
    subparser.set_defaults(func=validate_graph)


    subparser = subparsers.add_parser("merge_graphs")
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.add_argument("-g", "--graphs", nargs="+", required=True)
    subparser.set_defaults(func=merge_graphs_command)

    def make_variant_to_nodes(args):
        from .variant_to_nodes import VariantToNodes
        graph = Graph.from_file(args.graph)
        variants = VcfVariants.from_vcf(args.vcf, skip_index=True, dont_encode_chromosomes=True)
        variant_to_nodes = VariantToNodes.from_graph_and_variants(graph, variants)
        variant_to_nodes.to_file(args.out_file_name)
        logging.info("Wrote to file %s" % args.out_file_name)

    subparser = subparsers.add_parser("make_variant_to_nodes")
    subparser.add_argument("-g", "--graph", required=True)
    subparser.add_argument("-v", "--vcf", required=True)
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.set_defaults(func=make_variant_to_nodes)

    def make_node_to_variants(args):
        from .variant_to_nodes import NodeToVariants, VariantToNodes
        variant_to_nodes = VariantToNodes.from_file(args.variant_to_nodes)
        node_to_variants = NodeToVariants.from_variant_to_nodes(variant_to_nodes)
        node_to_variants.to_file(args.out_file_name)

    subparser = subparsers.add_parser("make_node_to_variants")
    subparser.add_argument("-v", "--variant_to_nodes", required=True)
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.set_defaults(func=make_node_to_variants)

    def create_coordinate_converter(args):
        from .coordinate_converter import CoordinateConverter
        converter = CoordinateConverter.from_graph(Graph.from_file(args.graph))
        converter.to_file(args.out_file_name)
        logging.info("Wrote to file %s" % args.out_file_name)

    subparser = subparsers.add_parser("create_coordinate_converter")
    subparser.add_argument("-g", "--graph", required=True)
    subparser.add_argument("-o", "--out_file_name", required=True)
    subparser.set_defaults(func=create_coordinate_converter)

    def intersect_vcfs(args):
        variants1 = VcfVariants.from_vcf(args.vcf1)
        variants2 = VcfVariants.from_vcf(args.vcf2)

        new = variants1.intersect(variants2)
        new.to_vcf_file(args.out_vcf, sample_name_output="")

    subparser = subparsers.add_parser("intersect_vcfs")
    subparser.add_argument("-a", "--vcf1", help="Vcf to intersect with other. Lines from this vcf will be kept")
    subparser.add_argument("-b", "--vcf2")
    subparser.add_argument("-o", "--out_vcf", help="Write resulting vcf to this file")
    subparser.set_defaults(func=intersect_vcfs)


    def make_numpy_variants(args):
        from .numpy_variants import NumpyVariants
        n = NumpyVariants.from_vcf(args.vcf)
        n.to_file(args.out_file_name)

    subparser = subparsers.add_parser("make_numpy_variants")
    subparser.add_argument("-v", "--vcf", required=True)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.set_defaults(func=make_numpy_variants)


    def make_position_id(args):
        from .position_id import PositionId
        graph = Graph.from_file(args.graph)
        position_id = PositionId.from_graph(graph)
        to_file(position_id, args.out_file_name)


    subparser = subparsers.add_parser("make_position_id")
    subparser.add_argument("-g", "--graph", required=True)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.set_defaults(func=make_position_id)


    def get_haplotype_sequence(args):
        # will traverse graph and follow these nodes
        # gets sequence for each chromosome in graph
        # writes fasta with sequences
        variant_nodes = np.load(args.nodes)
        nodes_to_follow = np.zeros(len(args.graph.nodes), dtype=np.uint8)
        nodes_to_follow[variant_nodes] = 1
        path_nodes, chromosome_indexes = traverse_graph_by_following_nodes(args.graph, nodes_to_follow, True)

        chromosome_indexes.append(len(path_nodes))

        chromosome_chunks = list(zip(chromosome_indexes[0:-1], chromosome_indexes[1:]))
        logging.info("Chromosome chunks: %s" % chromosome_chunks)

        chromosome_id = 1  # assume chromosomes are sorted
        fasta_lines = []
        coordinate_maps = {}
        refpos_to_node_maps = {}  # mapping from a ref pos in the haplotype coordinate space to node

        chromosome_index = 0
        for start_index, end_index in chromosome_chunks:
            nodes = path_nodes[start_index:end_index]
            fasta_lines.append(">" + str(chromosome_id) + "\n")
            haplotype_sequence = args.graph.get_nodes_sequence(nodes)
            fasta_lines.append(haplotype_sequence)
            fasta_lines.append("\n")

            # create a coordinate-map, a lookup from path pos to approx linear ref pos in graph
            coordinate_maps[str(chromosome_id)] = create_coordinate_map(nodes, args.graph, chromosome_index)

            refpos_to_node = np.zeros(len(haplotype_sequence), np.uint32)
            offsets = np.cumsum(args.graph.nodes[nodes])
            refpos_to_node[0] = nodes[0]
            refpos_to_node[offsets[:-1]] = nodes[1:]
            refpos_to_node = fill_zeros_with_last(refpos_to_node)
            refpos_to_node_maps[str(chromosome_id)] = refpos_to_node

            chromosome_index += 1
            chromosome_id += 1

        # write all sequences to fasta
        with open(args.out_file_name + ".fa", "w") as f:
            f.writelines(fasta_lines)
            logging.info("Wrote sequences to %s" % (args.out_file_name + ".fa"))

        # also write nodes in path to file
        np.save(args.out_file_name + ".nodes", path_nodes)
        logging.info("Wrote nodes in haplotype path to %s" % args.out_file_name + ".nodes.npy")

        to_file(coordinate_maps, args.out_file_name + ".coordinate_maps")
        logging.info("Wrote coordinate maps to %s" % args.out_file_name + ".coordinate_maps")

        to_file(refpos_to_node_maps, args.out_file_name + ".refpos_to_node")
        logging.info("Wrote refpos to node map to %s" % args.out_file_name + ".refpos_to_node")


    subparser = subparsers.add_parser("get_haplotype_sequence")
    subparser.add_argument("-n", "--nodes", required=True)
    subparser.add_argument("-g", "--graph", required=True, type=Graph.from_file)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.set_defaults(func=get_haplotype_sequence)

    def from_gfa(args):
        from .gfa import create_graph_from_gfa_file
        graph = create_graph_from_gfa_file(args.gfa)
        graph.to_file(args.out_file_name)

    subparser = subparsers.add_parser("from_gfa")
    subparser.add_argument("-g", "--gfa", required=True)
    subparser.add_argument("-o", "--out-file-name", required=True)
    subparser.set_defaults(func=from_gfa)


    def convert_gfa_ids_to_numeric_command(args):
        from .gfa import convert_gfa_ids_to_numeric
        convert_gfa_ids_to_numeric(args.gfa, args.out_base_name)


    subparser = subparsers.add_parser("convert_gfa_ids_to_numeric")
    subparser.add_argument("-g", "--gfa", required=True)
    subparser.add_argument("-o", "--out-base-name")
    subparser.set_defaults(func=convert_gfa_ids_to_numeric_command)

    if len(args) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(args)
    args.func(args)
    remove_shared_memory_in_session()




import logging


def traverse_graph_by_following_nodes(graph, follow_nodes, only_add_nodes_in_follow_nodes=False):
    assert type(follow_nodes) == set

    logging.info("N nodes in follow set: %d" % len(follow_nodes))

    nodes_found = []
    #linear_ref_nodes = graph.linear_ref_nodes()
    current_node = graph.get_first_node()
    logging.info("Starting traversing")
    chromosome_index = 0
    i = 0
    while True:
        if current_node in follow_nodes and only_add_nodes_in_follow_nodes:
            nodes_found.append(current_node)
        else:
            nodes_found.append(current_node)

        next_nodes = graph.get_edges(current_node)
        if len(next_nodes) == 0:
            # Check if we are at the end of a chromosome, if so, jump to next chromosome
            chromosome_index += 1
            if chromosome_index < len(graph.chromosome_start_nodes):
                next_nodes = [graph.chromosome_start_nodes[chromosome_index]]
                logging.info("Jumping to next chromosome with start node %d" % next_nodes[0])
            else:
                logging.info("Stopping at node %d. Did not find any next chromosome among nodes %s" % (current_node, graph.chromosome_start_nodes))
                break

        if i % 100000 == 0:
            logging.info("%i nodes traversed" % (i))

        next_node = None
        if len(next_nodes) == 1:
            next_node = next_nodes[0]
        else:
            for potential_next in next_nodes:
                if potential_next in follow_nodes or (next_node is None and graph.is_linear_ref_node(potential_next)):
                    next_node = potential_next

            # IF did not find any nodes, choose an empty node if there is one
            if next_node is None:
                next_node = [n for n in next_nodes if graph.get_node_size(n) == 0]
                assert len(next_node) == 1, "There are multiple empty nodes from node %d: %s" % (current_node, next_nodes)
                next_node = next_node[0]


        assert next_node is not None, "Could not find any new nodes from %d. Edges are %s" % (current_node, str(next_nodes))
        current_node = next_node
        i += 1

    if current_node in follow_nodes or not only_add_nodes_in_follow_nodes:
        nodes_found.append(current_node)

    return nodes_found



import logging
logging.basicConfig(level=logging.INFO)
from obgraph import Graph
import numpy as np


def test_from_dicts():
    g = Graph.from_dicts(
        {1: "ACTG", 2: "A", 3: "G", 4: "AAA"},
        {1: [2, 3],
         2: [4],
         3: [4]},
        [1, 2, 4]
    )

    assert g.get_node_size(1) == 4
    assert g.get_node_size(2) == 1
    assert g.get_node_size(3) == 1
    assert g.get_node_size(4) == 3

    assert g.get_node_sequence(2) == "A"

    assert list(g.get_edges(1)) == [2, 3]


    assert list(g.get_numeric_node_sequence(2)) == [0]

    print(g.get_numeric_node_sequence(np.array([1, 2, 3, 4])))
    assert list(g.get_numeric_node_sequences(np.array([1, 2, 3, 4]))) == [0, 1, 3, 2, 0, 2, 0, 0, 0]

    assert g.get_node_at_ref_offset(0) == 1
    assert g.get_node_at_ref_offset(3) == 1
    assert g.get_node_at_ref_offset(4) == 2
    assert g.get_node_at_ref_offset(5) == 4

    assert g.get_ref_offset_at_node(4) == 5
    assert g.get_ref_offset_at_node(2) == 4

    assert list(g.chromosome_start_nodes.values()) == [1]


def test_sparse_graph():
    g = Graph.from_dicts(
        {1: "AGGG", 4: "CACCT"},
        {1: [4]},
        [1, 4]
    )
    assert list(g.get_edges(1)) == [4]
    assert g.get_node_at_ref_offset(4) == 4
    assert g.get_nodes_sequence([1, 4]) == "AGGGCACCT"


def test2():
    g = Graph.from_dicts(
        {0: "ACT", 1: "A", 2: "", 3: "GGG"},
        {0: [1, 2], 1: [3], 2: [3]},
        [0, 3]
    )

    assert list(g.chromosome_start_nodes.values()) == [0]

    assert set(g.get_edges(0)) == set([1, 2])



def test_get_numeric_node_sequences_by_chromosome():
    g = Graph.from_dicts({
            1: "T", 2: "A", 1000: "A", 20: "ACAC", 23: "G", 50: "AAA", 51: "G"
        },
        {1: [2], 2: [1000], 1000: [20], 20: [23], 23: [50], 50: [51]},
        [1, 2, 1000, 20, 23, 50, 51],
        chromosome_start_nodes={"chromosomeI": 1, "chromosomeIV": 20, "chromosomeX": 50}
    )

    sequences = list(g.get_numeric_node_sequences_by_chromosomes([1, 2, 1000, 20, 23, 50, 51]))

    assert np.all(sequences[0] == [3, 0, 0])
    assert np.all(sequences[1] == [0, 1, 0, 1, 2])
    assert np.all(sequences[2] == [0, 0, 0, 2])





from obgraph.graph_merger import merge_graphs
from obgraph import Graph


def test_simple():

    graph1 = Graph.from_dicts(
        {1: "ACTG", 2: "A", 3: "C", 4: "ACT"},
        {1: [2, 3], 2: [4], 3: [4]},
        [1, 2, 4],
        chromosome_start_nodes={"chr1": 1}
    )

    graph2 = Graph.from_dicts(
        {1: "AAAA", 2: "A", 3: "C", 4: "ACT"},
        {1: [2, 3], 2: [4], 3: [4]},
        [1, 2, 4],
        chromosome_start_nodes={"chr2": 1}
    )

    merged_graph = merge_graphs([graph1, graph2])

    assert list(merged_graph.get_edges(1)) == [2, 3]
    assert list(merged_graph.get_edges(6)) == [7, 8]
    assert merged_graph.get_node_sequence(6) == "AAAA"
    assert merged_graph.get_node_sequence(7) == "A"
    assert merged_graph.get_node_sequence(8) == "C"

    assert merged_graph.get_node_at_ref_offset(0) == 1
    assert merged_graph.get_node_sequence(merged_graph.get_node_at_ref_offset(8)) == "AAAA"
    assert merged_graph.get_node_sequence(merged_graph.get_node_at_ref_offset(4)) == "A"
    assert len(merged_graph.get_edges(merged_graph.get_node_at_ref_offset(8))) == 2
    assert len(merged_graph.get_edges(merged_graph.get_node_at_ref_offset(11))) == 2

    assert merged_graph.get_ref_offset_at_node(6) == 8
    assert 7 in merged_graph.linear_ref_nodes()
    assert merged_graph.get_ref_offset_at_node(7) == 12

    assert merged_graph.chromosome_start_nodes == {"chr1": 1, "chr2": 6}

    merged_graph.to_file("merged_graph.npz")
    merged_graph2 = Graph.from_file("merged_graph.npz")

test_simple()


import numpy as np
import logging
from shared_memory_wrapper.shared_memory import to_shared_memory, from_shared_memory
from multiprocessing import Pool, Process
import time

class HaplotypeMatrix:
    properties = {"matrix"}
    def __init__(self, matrix=None):
        self.matrix = matrix

    def get_allele_frequency_for_nodes(self, nodes, node_to_variants, variant_to_nodes):
        # determine variant ids and which haplotype (ref or var) these nodes represent
        variant_ids = []
        haplotypes = []

        for node in nodes:
            variant_id = node_to_variants.get_variant_at_node(node)
            if variant_id is not None:
                variant_ids.append(variant_id)
                if variant_to_nodes.ref_nodes[variant_id] == node:
                    # ref node
                    haplotypes.append(1)
                else:
                    assert variant_to_nodes.var_nodes[variant_id] == node
                    haplotypes.append(2)

        if len(variant_ids) == 0:
            # no variants, meaning allele frequency is 1.0
            return 1.0

        return self.get_allele_frequency_for_variants_and_haplotypes(variant_ids, haplotypes)

    def get_allele_frequency_for_variants_and_haplotypes(self, variants, haplotypes):
        assert len(variants) == len(haplotypes)
        n_total_haplotypes = self.matrix.shape[0]
        # find the number of individual that have the given haplotypes at the given variants
        submatrix = self.matrix[:, variants]
        has_haplotypes = len(np.where(np.sum(submatrix == haplotypes, axis=1) == len(variants))[0])
        return has_haplotypes / n_total_haplotypes

    @classmethod
    def from_variants(cls, variants, n_individuals, n_variants, n_threads=10, chunk_size=10000):
        matrix = np.zeros((n_individuals*2, n_variants), dtype=np.uint8)
        matrix = cls(matrix)
        logging.info("Putting genotype matrix in shared memory")
        to_shared_memory(matrix, "haplotype_matrix")

        logging.info("Getting variant chunks")
        variant_chunks = variants.get_chunks(chunk_size=chunk_size)

        pool = Pool(n_threads)

        i = 0
        for result in pool.imap(HaplotypeMatrix.fill_shared_memory_matrix_with_variants, variant_chunks):
            i += 1
            logging.info("Done with %d variant chunks" % i)

        logging.info("Done with all variant chunks")
        matrix = from_shared_memory(HaplotypeMatrix, "haplotype_matrix")
        return cls(matrix.matrix)

    def get_alleles_with_haplotype_at_variant(self, variant_id, haplotype):
        assert haplotype [1, 2]
        return np.where(self.matrix[:, variant_id] == haplotype)[0]

    @staticmethod
    def fill_shared_memory_matrix_with_variants(variants):
        matrix = from_shared_memory(HaplotypeMatrix, "haplotype_matrix")

        for variant in variants:
            variant_number = variant.vcf_line_number
            if variant_number % 10000 == 0:
                logging.info("%d variants processeed" % variant_number)

            for individual_id, haplotypes in variant.get_individuals_and_numeric_haplotypes():
                matrix.matrix[individual_id*2, variant_number] = haplotypes[0]
                matrix.matrix[individual_id*2+1, variant_number] = haplotypes[1]

    def to_file(self, file_name):
        np.save(file_name, self.matrix)

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name)
        except FileNotFoundError:
            data = np.load(file_name + ".npy")

        return cls(data)



import numpy as np
import logging


class NumpyVariants:
    properties = {"header", "variants"}
    def __init__(self, header, variants):
        self.header = header
        self.variants = variants

    def to_file(self, file_name):
        np.savez(file_name,
            header=self.header,
            variants=self.variants
        )

    @classmethod
    def from_file(cls, file_name):
        data = np.load(file_name)
        return cls(data["header"], data["variants"])

    def to_vcf_with_genotypes(self, file_name, sample_name, genotypes, add_header_lines=None, ignore_homo_ref=False, add_genotype_likelyhoods=None):
        logging.info("Writing to file %s" % file_name)
        if ignore_homo_ref:
            logging.info("Will not write variants with genotype 0/0")

        if add_genotype_likelyhoods is not None:
            p = add_genotype_likelyhoods
            genotype_likelihoods = p * np.log10(np.e)

        with open(file_name, "w") as f:
            # f.write(self._header_lines)
            for header_line in self.header:  # last element is empty
                if header_line.startswith("#CHROM"):
                    if sample_name != "":
                        header_line = header_line.strip() + "\t" + sample_name + "\n"
                    if add_header_lines is not None:
                        for additional_header_line in add_header_lines:
                            f.writelines([additional_header_line + "\n"])
                f.writelines([header_line])

            lines = []
            for i, (variant, genotype) in enumerate(zip(self.variants, genotypes)):
                genotype = genotype.decode("utf-8")
                if ignore_homo_ref and genotype == "0/0":
                    continue

                variant = variant.decode("utf-8").strip()
                if i % 1000000 == 0:
                    logging.info("%d variants written to file." % i)

                line = "%s\t%s\n" % (variant, genotype)
                if add_genotype_likelyhoods is not None:
                    """
                    #likelyhoods = add_genotype_likelyhoods[i]
                    phred_likelyhoods = [
                        #float(round(np.maximum(0, np.minimum(255, -prob * np.log10(np.e))), 4))
                        int(np.minimum(2550000000000, -10 * prob * np.log10(np.e)))
                        for prob in likelyhoods
                    ]
                    """
                    gl_str = ",".join(str(round(p, 4)) for p in genotype_likelihoods[i])
                    #print(likelyhoods, phred_likelyhoods, np.exp(likelyhoods))
                    line = "%s:GL\t%s:%s\n" % (variant, genotype, gl_str)
                    #print(line)

                lines.append(line)

            f.writelines(lines)

    @classmethod
    def from_vcf(cls, vcf_file_name):
        header = []
        variants = []
        with open(vcf_file_name, "r") as f:
            for i, line in enumerate(f):
                #line_decoded = line.decode("utf-8")
                if line.startswith("#"):
                    header.append(line)
                else:
                    # add byte line
                    # remove info field to save space
                    l = line.split()
                    l[7] = "."
                    line = "\t".join(l)
                    variants.append(str.encode(line))  # add as byte

                if i % 10000 == 0:
                    logging.info("%d variants processed" % i)

        variants = np.array(variants)
        logging.info("Variants type: %s" % variants.dtype)
        return cls(np.array(header, dtype=str), variants)





import numpy as np
import logging
from .cython_traversing import fill_zeros_increasingly
import resource


def encode_chromosome(chromosome):
    if chromosome.startswith("chr"):
        return encode_chromosome(chromosome.replace("chr", ""))

    if chromosome.upper() == "X":
        return 23
    elif chromosome.upper() == "Y":
        return 24
    else:
        return int(chromosome)


def phased_genotype_matrix_to_haplotype_matrix(genotype_matrix):
    haplotype_matrix = np.zeros((genotype_matrix.shape[0], genotype_matrix.shape[1]*2), dtype=genotype_matrix.dtype)
    haplotype_matrix[:,::2] = (genotype_matrix // 2) == 1
    haplotype_matrix[:,1::2] = (genotype_matrix % 2) == 1
    return haplotype_matrix


def add_indel_dummy_nodes(graph):
    from .graph import Graph
    node_ids, node_sequences, node_sizes, from_nodes, to_nodes, linear_ref_nodes, chromosome_start_nodes = graph.get_flat_graph()

    linear_ref_set = set(linear_ref_nodes)

    node_counter = node_ids[-1] + 1

    change_edges = {}
    reverse_edges = graph.get_reverse_edges_dict()
    # traverse graph, detect indels
    for node in node_ids:
        if node in linear_ref_set:
            edges = graph.get_edges(node)
            if len(edges) > 1:
                linear_ref_pos_end_of_node = graph.get_ref_offset_at_node(node) + graph.get_node_size(node)

                # Find all edges going to linear ref node where
                for next_node in edges:
                    # deletions and insertions

                    # New smarter detection that handles cases with overlapping variant
                    # if edge goes from linear ref node to another linear ref node where next linear ref node ref pos is not directly after
                    #   this must be a deletion, add a dummy node
                    # if edge goes directly to another linear ref node that follows on the linear ref and there are other edges that do not go to linear ref nodes and these go to the same linear ref node that is the next
                    #  there must be an insertion, this edge should be replace
                    #   this solution does not support insertions with more than one node in the insertion, but that does not happen??
                    next_node_linear_ref_pos = graph.get_ref_offset_at_node(next_node)
                    #logging.info("Checking edge from %d to %d. Linear ref end: %d. Next linear ref pos: %d" % (node, next_node, linear_ref_pos_end_of_node, next_node_linear_ref_pos))
                    """
                    if (next_node in linear_ref_set and next_node_linear_ref_pos > linear_ref_pos_end_of_node) \
                            or (next_node in linear_ref_set and next_node_linear_ref_pos == linear_ref_pos_end_of_node and \
                                len([other_next for other_next in edges if other_next != next_node and other_next not in linear_ref_set and next_node in graph.get_edges(other_next)]) > 0):
                        #logging.info("Adding edge between node %d and %d" % (node, next_node))
                        node_ids.append(node_counter)
                        node_sequences.append([""])
                        node_sizes.append(0)
                        from_nodes.append(node)
                        to_nodes.append(node_counter)
                        change_edges[(node, next_node)] = (node_counter, next_node)
                        node_counter += 1
                    """

                    # Alternative method that also deals with cases where an insertion node has multiple edges in
                    # insertions
                    if next_node in linear_ref_set and next_node_linear_ref_pos == linear_ref_pos_end_of_node:
                        #print("Checking insertion between %d and %d" % (node, next_node))
                        insertion_nodes = [n for n in edges if n != next_node and n not in linear_ref_set and next_node in graph.get_edges(n)]
                        #print("  Insertion nodes: %s" % insertion_nodes)
                        for insertion_node in insertion_nodes:
                            # There might be multiple insertions we assume
                            # For every node having an edge into the insertion node, we want to add a dummy node between that node and the next_node (next node on lienar ref)

                            nodes_in = reverse_edges[insertion_node]
                            #print("  Nodes going in to %d: %s" % (insertion_node, nodes_in))
                            if len(nodes_in) > 0:
                                node_ids.append(node_counter)
                                node_sequences.append([""])
                                node_sizes.append(0)
                                # Add edge from dummy node to next
                                from_nodes.append(node_counter)
                                to_nodes.append(next_node)
                                # Change all old edges from previous node to next_node to go to dummy node
                                for node_in in nodes_in:
                                    #print("   Changing node from %d-%d to %d-%d" % (node_in, next_node, node_in, node_counter))
                                    change_edges[(node_in, next_node)] = (node_in, node_counter)
                                node_counter += 1

                    # Deletion
                    if next_node in linear_ref_set and next_node_linear_ref_pos > linear_ref_pos_end_of_node:
                        #print("CHecking deletion from %d to %d" % (node, next_node))
                        # Insertion nodes can only be linear reference nodes that are before the next node
                        insertion_nodes = [n for n in edges if n != next_node and n in linear_ref_set and graph.get_ref_offset_at_node(n) < graph.get_ref_offset_at_node(next_node)]
                        #print("  Insertion nodes on linear ref: %s" % insertion_nodes)
                        for insertion_node in insertion_nodes:
                            # There cannot be multiple deletions to the same other ref node? Still looping
                            # For every node having an edge into the insertion node, we want to add a dummy node between that node and the next_node (next node on lienar ref)

                            nodes_in = reverse_edges[insertion_node]
                            if len(nodes_in) > 0:
                                node_ids.append(node_counter)
                                node_sequences.append([""])
                                node_sizes.append(0)
                                # Add edge from dummy node to next
                                from_nodes.append(node_counter)
                                to_nodes.append(next_node)
                                #print("   Adding new dummy node %d and edge from %d to %d" % (node_counter, node_counter, next_node))
                                # Change all old edges from previous node to next_node to go to dummy node
                                for node_in in nodes_in:
                                    #print("   Changing edte %d-%d to %d-%d" % (node_in, next_node, node_in, node_counter))
                                    change_edges[(node_in, next_node)] = (node_in, node_counter)
                                node_counter += 1


                    """
                    if next_node in linear_ref_set and len([other_next for other_next in edges if other_next != next_node and next_node in graph.get_edges(other_next)]) > 0:
                        node_ids.append(node_counter)
                        node_sequences.append([""])
                        node_sizes.append(0)
                        from_nodes.append(node)
                        to_nodes.append(node_counter)
                        change_edges[(node, next_node)] = (node_counter, next_node)
                        node_counter += 1
                    """

    # Change edges
    for i in range(len(from_nodes)):
        from_node = from_nodes[i]
        to_node = to_nodes[i]
        if (from_node, to_node) in change_edges:
            from_nodes[i] = change_edges[(from_node, to_node)][0]
            to_nodes[i] = change_edges[(from_node, to_node)][1]

    return Graph.from_flat_nodes_and_edges(np.array(node_ids), node_sequences, np.array(node_sizes), np.array(from_nodes), np.array(to_nodes), linear_ref_nodes, chromosome_start_nodes)


def get_number_of_variants_and_individuals_from_vcf(file_name):
    raise NotImplementedError("")
    file = open(file_name)
    for line in file:
        continue


def fill_zeros_with_last(arr, initial=0):
    ind = np.nonzero(arr)[0]
    cnt = np.cumsum(np.array(arr, dtype=bool))
    return np.where(cnt, arr[ind[cnt-1]], initial)


def create_coordinate_map(path_nodes, graph, chromosome_index):
    path_node_sizes = graph.nodes[path_nodes]
    path_offsets = np.concatenate([[0], np.cumsum(path_node_sizes)[:-1]]).astype(int)
    linear_ref_offsets = graph.node_to_ref_offset[path_nodes] - graph.get_ref_offset_at_node(list(graph.chromosome_start_nodes.values())[chromosome_index])

    print(path_offsets)
    print(linear_ref_offsets)

    # lookup is from path_offsets to linear_ref_offsets
    lookup = np.zeros(int(path_offsets[-1])+1, dtype=np.int64)
    lookup[path_offsets] = linear_ref_offsets
    print(lookup)
    fill_zeros_increasingly(lookup)

    """
    # should fill zeros with 1 more than last, if not large nodes will not give accurate mapping
    lookup = fill_zeros_with_last(lookup)

    # Add node offsets
    path_node_to_dist = np.zeros(int(path_offsets[-1])+1)
    path_node_to_dist[path_offsets] = path_offsets
    path_node_to_dist = fill_zeros_with_last(path_node_to_dist)
    print("Path node to dist: %s" % path_node_to_dist)

    node_offsets = np.arange(0, len(path_node_to_dist)) - path_node_to_dist
    print("Lookup before adding: %s" % lookup)
    lookup += node_offsets

    """
    print("Final lookup: %s" % lookup)
    return lookup



def log_memory_usage_now(logplace=""):
    memory = int(resource.getrusage(resource.RUSAGE_SELF).ru_maxrss) / 1000000
    logging.info("Memory usage (%s): %.4f GB" % (logplace, memory))





from obgraph.util import create_coordinate_map
from obgraph import Graph
import numpy as np
from obgraph.util import phased_genotype_matrix_to_haplotype_matrix


def test_create_coordinate_map():

    g = Graph.from_dicts(
        {1: "AAAA", 2: "G", 3: "GGGGGGG", 4: "ACA", 5: "C"},
        {1: [2, 3], 2: [4], 3: [4], 4: [5]},
        [1, 2, 4, 5]
    )

    path = np.array([1, 3, 4, 5])

    m = create_coordinate_map(path, g, chromosome_index=0)

    assert m[0] == 0
    assert m[11] == 5
    assert m[14] == 8


def test_phased_genotype_matrix_to_haplotype_matrix():
    matrix = np.array([
        [0, 1, 2, 3],
        [3, 2, 1, 0]
    ])

    true = np.array([
        [0, 0, 0, 1, 1, 0, 1, 1],
        [1, 1, 1, 0, 0, 1, 0, 0]
    ])


    assert np.all(
        phased_genotype_matrix_to_haplotype_matrix(matrix) == true
    )

test_create_coordinate_map()




import numpy as np
import logging
import time
from .graph import VariantNotFoundException
from shared_memory_wrapper.shared_memory import to_shared_memory, from_shared_memory, SingleSharedArray
from multiprocessing import Pool, Process
from .variant_to_nodes import VariantToNodes
import itertools
from dataclasses import dataclass
from .haplotype_nodes import get_variant_matrix_as_chunks_with_variant_ids
import bionumpy as bnp
import gzip



@dataclass
class PhasedGenotypeMatrix:
    matrix: np.ndarray

    @classmethod
    def from_vcf(cls, vcf_file_name, n_variants, n_individuals):
        t = time.perf_counter()
        matrix = np.zeros((n_variants, n_individuals), dtype=np.uint8)
        logging.info("Matrix size: %.2f GB" % (matrix.nbytes/1000000000))
        file = bnp.open(vcf_file_name, buffer_type=bnp.PhasedVCFMatrixBuffer)
        #for start_variant, end_variant, variants in get_variant_matrix_as_chunks_with_variant_ids(vcf_file_name):
        pos = 0
        for chunk in file.read_chunk(min_chunk_size=15000000):
            genotypes = chunk.genotypes.raw()
            matrix[pos:pos+chunk.genotypes.shape[0],:] = genotypes
            pos += genotypes.shape[0]
            logging.info("Processed %d variants. Done with %d variants in %.4f sec" % (
                len(genotypes),
                pos,
                time.perf_counter()-t
            ))

        return cls(matrix)

    @classmethod
    def from_txt(cls, txt_file_name, n_variants, n_individuals):
        t0 = time.perf_counter()
        data_size = n_individuals * n_variants * 2
        logging.info("N individuals: %d" % n_individuals)
        logging.info("N variants: %d" % n_variants)
        logging.info("Data size is %d" % data_size)
        data = np.zeros(data_size, dtype=np.uint8)
        logging.info("Made data array")
        f = gzip.open(txt_file_name, "rb")
        f.readinto(data)

        logging.info("Read data. Time spent: %.3f sec " % (time.perf_counter() - t0))

        # replace "." (unknown genotype)
        is_unknown = data == 46
        logging.info("Number of unknown genotypes (these will be replaced by reference): %d/%d" % (np.sum(is_unknown), len(data)))
        data[is_unknown] = 48

        assert np.max(data) <= 49, "There are values larger than 49 (1) in data. Invalid VCF txt format " + str(np.where(data > 49))
        assert np.min(data) >= 48, "There are values lower than 48 (0) in data (%d). Invalid VCF txt format %s"  % (np.min(data), str(np.where(data < 48)))

        logging.info("Done validating")

        data = data - 48
        genotypes = data[::2] * 2 + data[1::2]
        logging.info("Made genotypes. Time spent: %.3f sec" % (time.perf_counter() - t0))

        assert np.max(data) <= 3
        assert np.min(data) >= 0

        return cls(genotypes.reshape((n_variants, n_individuals)))


class MostSimilarVariantLookup:
    properties = {"lookup_array", "prob_same_genotype"}
    def __init__(self, lookup_array=None, prob_same_genotype=None):
        self.lookup_array = lookup_array
        self.prob_same_genotype = prob_same_genotype

    def get_most_similar_variant(self, variant_id):
        return self.lookup_array[variant_id]

    def prob_of_having_the_same_genotype_as_most_similar(self, variant_id):
        return self.prob_same_genotype[variant_id]

    def to_file(self, file_name):
        np.savez(file_name, lookup_array=self.lookup_array, prob_same_genotype=self.prob_same_genotype)

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name)
        except FileNotFoundError:
            data = np.load(file_name + ".npz")

        return cls(data["lookup_array"], data["prob_same_genotype"])


class GenotypeTransitionProbabilities:
    properties = {"matrix"}
    def __init__(self, matrix=None):
        self.matrix = matrix

    def to_file(self, file_name):
        np.save(file_name, self.matrix)

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name)
        except FileNotFoundError:
            data = np.load(file_name + ".npy")

        return cls(data)

    def get_transition_probabilities(self, variant_id, from_genotype):
        assert from_genotype in [0, 1, 2]
        matrix_index = from_genotype * 3
        return self.matrix[matrix_index:matrix_index+3, variant_id]

    def get_transition_probability(self, variant_id, from_genotype, to_genotype):
        assert from_genotype in [0, 1, 2]
        assert to_genotype in [1, 2, 0]
        matrix_index = from_genotype*3 + to_genotype
        return self.matrix[matrix_index, variant_id]

    @staticmethod
    def create_using_shared_memory(variant_interval):
        from_variant, to_variant = variant_interval
        logging.info("Creating on interval %d-%d" % (from_variant, to_variant))

        genotype_matrix = from_shared_memory(GenotypeMatrix, "genotype_matrix_shared")
        most_similar_variants = from_shared_memory(MostSimilarVariantLookup, "most_similar_variants_shared")
        probability_matrix = from_shared_memory(GenotypeTransitionProbabilities, "transition_probs_shared")
        matrix = probability_matrix.matrix

        for variant_id in range(from_variant, to_variant):
            if variant_id % 100000 == 0:
                logging.info("%d/%d variants processed" % (variant_id-from_variant, to_variant-from_variant))

            most_similar_variant = most_similar_variants.get_most_similar_variant(variant_id)
            transition_probs = genotype_matrix.get_transitions_probs_between_variants(most_similar_variant, variant_id)
            matrix[:, variant_id] = transition_probs

    @classmethod
    def from_most_similar_variants_and_matrix(cls, most_similar_variants, genotype_matrix, n_threads=10):
        n_variants = len(most_similar_variants.lookup_array)
        matrix = cls(np.zeros((9, n_variants), dtype=np.float))
        to_shared_memory(matrix, "transition_probs_shared")
        to_shared_memory(genotype_matrix, "genotype_matrix_shared")
        to_shared_memory(most_similar_variants, "most_similar_variants_shared")

        intervals = [int(i) for i in np.linspace(0, n_variants, n_threads)]
        variant_intervals = [(from_id, to_id) for from_id, to_id in zip(intervals[0:-1], intervals[1:])]
        logging.info("Will analyse intervals: %s" % variant_intervals)

        pool = Pool(n_threads)

        for result in pool.imap(GenotypeTransitionProbabilities.create_using_shared_memory, variant_intervals):
            logging.info("Done with one job")

        return from_shared_memory(GenotypeTransitionProbabilities, "transition_probs_shared")


class GenotypeFrequencies:
    properties = {"homo_ref", "homo_alt", "hetero"}
    def __init__(self, homo_ref=None, homo_alt=None, hetero=None):
        self.homo_ref = homo_ref
        self.homo_alt = homo_alt
        self.hetero = hetero

    @staticmethod
    def create_using_shared_memory(variant_interval):
        from_variant, to_variant = variant_interval
        logging.info("Creating on interval %d-%d" % (from_variant, to_variant))

        genotype_matrix = from_shared_memory(GenotypeMatrix, "genotype_matrix_shared_for_frequencies")
        genotype_frequencies = from_shared_memory(GenotypeFrequencies, "genotype_frequencies_shared")

        n_variants = genotype_matrix.matrix.shape[1]
        n_individuals = len(np.where(genotype_matrix.matrix[:,0])[0] != 0)  # can be zeros for non-individuals, so all non-zero is an individual
        # Less memory hungry, but slower

        for numeric_genotype, array in zip([1, 2, 3], [genotype_frequencies.homo_ref, genotype_frequencies.homo_alt, genotype_frequencies.hetero]):
            logging.info("Finding for genotype %d" % numeric_genotype)
            prev_time = time.time()
            for variant_id in range(from_variant, to_variant):
                if variant_id % 100000 == 0:
                    logging.info("%d/%d variants processed (genotype now is %d). Prev 100k processed in %.3f s" % (variant_id-from_variant, to_variant-from_variant, numeric_genotype, time.time()-prev_time))
                    prev_time = time.time()

                array[variant_id] = len(np.where(genotype_matrix.matrix[:, variant_id] == numeric_genotype)[0]) / n_individuals

    @classmethod
    def create_naive_from_vcf_af_field(cls, variants):
        homo_ref = []
        homo_alt = []
        hetero = []
        for variant in variants:
            allele_frequency = variant.get_variant_allele_frequency()
            homo_alt.append(allele_frequency**2)
            homo_ref.append((1-allele_frequency)**2)
            hetero.append(1 - allele_frequency**2 - (1-allele_frequency)**2)

        return cls(np.array(homo_ref), np.array(homo_alt), np.array(hetero))

    @classmethod
    def from_genotype_matrix(cls, genotype_matrix, n_threads=10):
        to_shared_memory(genotype_matrix, "genotype_matrix_shared_for_frequencies")

        n_variants = genotype_matrix.matrix.shape[1]
        n_individuals = len(np.where(genotype_matrix.matrix[:,0])[0] != 0)  # can be zeros for non-individuals, so all non-zero is an individual
        logging.info("Assumes there are %d individuals and %d variants" % (n_individuals, n_variants))
        data = {1: np.zeros(n_variants, dtype=float), 2: np.zeros(n_variants, dtype=float), 3: np.zeros(n_variants, dtype=float)}
        genotype_frequences = cls(data[1], data[2], data[3])
        to_shared_memory(genotype_frequences, "genotype_frequencies_shared")

        intervals = [int(i) for i in np.linspace(0, n_variants, n_threads)]
        variant_intervals = [(from_id, to_id) for from_id, to_id in zip(intervals[0:-1], intervals[1:])]
        logging.info("Will analyse intervals: %s" % variant_intervals)

        pool = Pool(n_threads)

        for result in pool.imap(GenotypeFrequencies.create_using_shared_memory, variant_intervals):
            logging.info("Done with one job")

        """
        for numeric_genotype, array in data.items():
            logging.info("Finding for genotype %d" % numeric_genotype)
            # the second index from np where gives the columns that have a hit, every column 1 time for each hit
            column_hits = np.where(genotype_matrix.matrix == numeric_genotype)[1]
            logging.info("Making frequencies")
            unique_columns, n_hits_per_column = np.unique(column_hits, return_counts=True)
            data[numeric_genotype][unique_columns] = n_hits_per_column / n_individuals
        """
        """
        # Less memory hungry, but slower
        for numeric_genotype, array in data.items():
            logging.info("Finding for genotype %d" % numeric_genotype)
            for variant_id in range(n_variants):
                if variant_id % 10000 == 0:
                    logging.info("%d variants processed" % variant_id)

                array[variant_id] = len(np.where(genotype_matrix.matrix[:,variant_id] == numeric_genotype)[0]) / n_individuals
        """
        return from_shared_memory(GenotypeFrequencies, "genotype_frequencies_shared")
        #return cls(data[1], data[2], data[3])

    def get_frequencies_for_variant(self, variant_id):
        return self.homo_ref[variant_id], self.homo_alt[variant_id], self.hetero[variant_id]

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name)
        except FileNotFoundError:
            data = np.load(file_name + ".npz")

        return cls(data["homo_ref"], data["homo_alt"], data["hetero"])

    def to_file(self, file_name):
        np.savez(file_name, homo_ref=self.homo_ref, homo_alt=self.homo_alt, hetero=self.hetero)


class GenotypeMatrixAnalyser:
    def __init__(self, genotype_matrix, whitelist_array=None):
        self.matrix = genotype_matrix
        self._whitelist_array = whitelist_array

    @staticmethod
    def analyse_variants_on_shared_memody(variant_interval):
        from_id, to_id = variant_interval
        if from_id == 0:
            from_id = 1
        logging.info("Analysing variant %d to %d in one job" % (from_id, to_id))
        whitelist_array = from_shared_memory(SingleSharedArray, "whitelist_variants").array
        matrix = from_shared_memory(GenotypeMatrix, "genotype_matrix")
        lookup = from_shared_memory(MostSimilarVariantLookup, "most_similar_variant_lookup")
        n_individuals = matrix.matrix.shape[0]
        prev_time = time.time()
        for i, variant_id in enumerate(range(from_id, to_id)):
            if i % 50000 == 0 and i > 0:
                logging.info("%d/%d variants analysed (last 50k analysed in %.3f s)" % (i, to_id-from_id, time.time()-prev_time))
                prev_time = time.time()

            most_similar, score = matrix.get_most_similar_previous_variant(variant_id, whitelist_array)
            #logging.info("Most similar to %d is %d with score %d. Genotype distribution: %s" % (variant_id, most_similar, score, np.unique(self.matrix[:,variant_id], return_counts=True)))
            lookup.lookup_array[variant_id] = most_similar
            lookup.prob_same_genotype[variant_id] = score / n_individuals


    def analyse(self, n_threads=10):
        n_variants = self.matrix.matrix.shape[1]
        n_individuals = self.matrix.matrix.shape[0]

        most_similar_lookup = np.zeros(n_variants, dtype=np.uint32)
        prob_same_genotype = np.zeros(n_variants, dtype=np.float)

        to_shared_memory(SingleSharedArray(self._whitelist_array), "whitelist_variants")

        lookup = MostSimilarVariantLookup(most_similar_lookup, prob_same_genotype)
        to_shared_memory(self.matrix, "genotype_matrix")
        to_shared_memory(lookup, "most_similar_variant_lookup")

        intervals = [int(i) for i in np.linspace(0, n_variants, n_threads)]
        variant_intervals = [(from_id, to_id) for from_id, to_id in zip(intervals[0:-1], intervals[1:])]
        logging.info("Will analyse intervals: %s" % variant_intervals)

        pool = Pool(n_threads)

        for result in pool.imap(GenotypeMatrixAnalyser.analyse_variants_on_shared_memody, variant_intervals):
            logging.info("Done with one job")

        lookup = from_shared_memory(MostSimilarVariantLookup, "most_similar_variant_lookup")

        return lookup



class GenotypeMatrix:
    properties = {"matrix"}

    def __init__(self, matrix=None):
        if matrix is not None:
            logging.info("Type of matrix: %s" % matrix.dtype)
            logging.info("Size of matrix: %3.f MB" % (int(matrix.nbytes)/1000000))
        self.matrix = matrix

    @classmethod
    def from_variants(cls, variants, n_individuals=None, n_variants=None, n_threads=10, chunk_size=10000):
        shared_memory_unique_id = str(np.random.randint(0, 10e15))

        if n_variants is None:
            logging.warning("Finding n variants and n individuals by counting, may be slow")
            n_variants = variants.n_variants()
            n_individuals = variants.n_individuals()

        matrix = np.zeros((n_individuals, n_variants), dtype=np.uint8) + 4  # 4 is unknown genotype
        matrix = cls(matrix)
        logging.info("Putting genotype matrix in shared memory")
        to_shared_memory(matrix, "genotype_matrix"+shared_memory_unique_id)

        logging.info("Getting variant chunks")
        variant_chunks = variants.get_chunks(chunk_size=chunk_size)

        pool = Pool(n_threads)

        i = 0
        for result in pool.imap(GenotypeMatrix.fill_shared_memory_matrix_with_variants, zip(variant_chunks, itertools.repeat(shared_memory_unique_id))):
            i += 1
            logging.info("Done with %d variant chunks" % i)

        logging.info("Done with all variant chunks")
        matrix = from_shared_memory(GenotypeMatrix, "genotype_matrix"+shared_memory_unique_id)
        return cls(matrix.matrix)

    def convert_to_other_format(self):
        genotype_matrix = self.matrix.transpose()

        # genotypes are 1, 2, 3 (0 for unknown, 1 for homo ref, 2 for homo alt and 3 for hetero), we want 0, 1, 2 for homo alt, hetero, homo ref
        logging.info("Converting genotype matrix with size %s" % str(self.matrix.shape))
        # 0, 1 => 2
        # 2 => 0
        # 3 => 1
        new_genotype_matrix = np.zeros_like(genotype_matrix, dtype=np.int8)
        idx = np.where(genotype_matrix == 0)[0]
        #logging.info("Index size: %d" % (int(idx.nbytes)/1000000))
        new_genotype_matrix[np.where(genotype_matrix == 0)] = -1
        logging.info("done 1/4")
        new_genotype_matrix[np.where(genotype_matrix == 1)] = 0
        logging.info("done 2/4")
        new_genotype_matrix[np.where(genotype_matrix == 2)] = 2
        logging.info("done 3/4")
        new_genotype_matrix[np.where(genotype_matrix == 3)] = 1
        logging.info("done 4/4")
        return GenotypeMatrix(new_genotype_matrix)

    def get_transitions_probs_between_variants(self, from_variant_id, to_variant_id):
        matrix = self.matrix
        probs = np.zeros(3*3)  # the 9 probs will be stored in a flat array
        for from_genotype in [0, 1, 2]:
            for to_genotype in [0, 1, 2]:
                genotype_indexes = np.where(matrix[:, from_variant_id] == from_genotype)[0]
                n_total = len(genotype_indexes)
                n_kept = len(np.where(matrix[genotype_indexes, to_variant_id] == to_genotype)[0])

                if n_total == 0:
                    prob = 0.0
                else:
                    prob = n_kept / n_total

                probs_index = 3*from_genotype + to_genotype
                probs[probs_index] = prob

        return probs

    def get_individuals_with_genotype_at_variant(self, variant_id, genotype):
        assert genotype in [0, 1, 2]
        return np.where(self.matrix[:, variant_id] == genotype)[0]

    def get_transition_prob_from_single_to_multiple_variants(self, from_variant_id, from_genotype, to_variant_ids_and_genotypes):

        # init to individuals at from_variant
        shared_individuals = set(self.get_individuals_with_genotype_at_variant(from_variant_id, from_genotype))
        n_at_from_variant = len(shared_individuals)

        if n_at_from_variant == 0:
            return 0

        for to_variant, to_genotype in to_variant_ids_and_genotypes:
            individuals = self.get_individuals_with_genotype_at_variant(to_variant, to_genotype)
            shared_individuals = shared_individuals.intersection(individuals)
            if len(shared_individuals) == 0:
                break

        n_left = len(shared_individuals)

        return n_left / n_at_from_variant

    def get_transition_prob(self, from_variant_id, to_variant_id, from_genotype, to_genotype):
        assert from_genotype in [0, 1, 2]
        assert to_genotype in [0, 1, 2]
        matrix = self.matrix
        genotype_indexes = np.where(matrix[:, from_variant_id] == from_genotype)[0]
        n_total = len(genotype_indexes)
        n_kept = len(np.where(matrix[genotype_indexes, to_variant_id] == to_genotype)[0])

        if n_total == 0:
            prob = 0.0
        else:
            prob = n_kept / n_total

        return prob

    def get_most_similar_previous_variant(self, variant_id, whitelist_array=None, window=1000):
        matrix = self.matrix
        #variant_genotypes = matrix[:,variant_id]
        #print("Variant genotypes: %s" % variant_genotypes)
        submatrix_start = 0
        submatrix_end = variant_id + 1
        # Only look at 1000 variants back
        if variant_id > window:
            submatrix_start = variant_id - window

        if submatrix_end > matrix.shape[1]:
            submatrix_end = matrix.shape[1]

        submatrix = matrix[:,submatrix_start:submatrix_end]
        #print("Submatrix: %s" % submatrix)
        similarity_scores = np.sum(
            submatrix.transpose() == matrix[:,variant_id],
            axis=1
        )
        #print(similarity_scores)
        # set score for self to -1 to ignore
        similarity_scores[variant_id-submatrix_start] = -1

        if whitelist_array is not None:
            # all variants not marked by 1 in whitelist array should not be chosen, give these low similarity score
            similarity_scores[whitelist_array[submatrix_start:submatrix_end] == 0] = -1

        most_similar = np.argmax(similarity_scores) + submatrix_start
        value = similarity_scores[most_similar - submatrix_start]
        return most_similar, value

    @staticmethod
    def fill_shared_memory_matrix_with_variants(data):
        variants, shared_memory_unique_id = data
        logging.info("Handling subset, using encoding version 2")
        matrix = from_shared_memory(GenotypeMatrix, "genotype_matrix"+shared_memory_unique_id)
        n_individuals = matrix.matrix.shape[0]
        logging.info("There are %d individuals" % n_individuals)

        for variant in variants:
            variant_number = variant.vcf_line_number
            if variant_number % 10000 == 0:
                logging.info("%d variants processeed" % variant_number)

            for individual_id, genotype in variant.get_individuals_and_numeric_genotypes(encoding_version="2"):
                if individual_id >= n_individuals:
                    break

                matrix.matrix[individual_id, variant_number] = genotype

    @classmethod
    def from_nodes_to_haplotypes_and_variants(cls, nodes_to_haplotypes, variants, graph, n_individuals):

        n_variants = len(variants)

        matrix = np.zeros((n_individuals, n_variants), dtype=np.uint8)

        for variant_number, variant in enumerate(variants):
            if variant_number % 100 == 0:
                logging.info("%d variants processeed" % variant_number)

            try:
                reference_node, variant_node = graph.get_variant_nodes(variant)
            except VariantNotFoundException:
                continue

            for individual_id in nodes_to_haplotypes.get_individuals_having_node_pair(reference_node, reference_node):
                matrix[individual_id, variant_number] = 1

            for individual_id in nodes_to_haplotypes.get_individuals_having_node_pair(variant_node, variant_node):
                matrix[individual_id, variant_number] = 2

            for individual_id in nodes_to_haplotypes.get_individuals_having_node_pair(reference_node, variant_node):
                matrix[individual_id, variant_number] = 3

        return cls(matrix)

    def to_file(self, file_name):
        np.save(file_name, self.matrix)

    @classmethod
    def from_file(cls, file_name):
        try:
            data = np.load(file_name)
        except FileNotFoundError:
            data = np.load(file_name + ".npy")

        return cls(data)





class NeighbourHaplotypeNodes:
    def __init__(self, index):
        self._index = index

    @classmethod
    def from_variants(cls, variants, variant_to_nodes):
        for variant_index in range(len(variants)):
            # Iterate last 4 variants,


